#!/usr/bin/env python
# coding: utf-8

import os

exists = os.path.exists
abspatch = os.path.abspath
dirname = os.path.dirname


'''
Old, from devapp/arch/axc2, will need a few tools later:
"""
# OS Tools

Supporting various operations on the shell, like mkdir, chroot...

"""
import os, sys, stat, time, errno, re, inspect, platform
from os.path import exists
import json, socket, stat
import subprocess as sp
from . import sh
from axc_app.cli import say, die, dbg, quiet, warn, add_indent, pop_indent
from axc_app.app import app_dir, call, base_pid
from time import sleep, time
from .os_tools import j, exists, env, ls
import tempfile, uuid
import shutil
from fnmatch import fnmatch
from functools import partial
from axc_app.tools import cast_list, is_str, str8

join, isdir, islink = os.path.join, os.path.isdir, os.path.islink

piped_popen = partial(
    sp.Popen, shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE
)

XC = '/xc'


def term_width():
    width = 0
    try:
        import struct, fcntl, termios

        s = struct.pack('HHHH', 0, 0, 0, 0)
        x = fcntl.ioctl(1, termios.TIOCGWINSZ, s)
        width = struct.unpack('HHHH', x)[1]
    except IOError:
        pass
    return width


def do_logo():
    """ show the AXC logo"""
    c = lambda c, m='38': '\\033[%s;5;%sm' % (m, c)
    with open(j(app_dir, 'logo')) as f:
        lg = f.read() + '\n'
    try:
        cols = term_width()
    except:
        print('welcome to axc2')
        return
    if not cols:
        return
    # lg = '\n'.join([l.ljust(cols) for l in lg.splitlines()])
    os.system("echo -e '\\e[48;5;233m'")
    for L in ('B', '1', '4', '5', '6', '7'):
        lg = lg.replace(L, L + L + L)

    for k, v in list(
        {
            '\n': '\n' + c(232),
            'BBB': c(233),
            '111': c(236),
            '444': c(82),
            '555': c(236),
            '666': c(235),
            '777': c(234),
            '=': '■',
        }.items()
    ):
        lg = lg.replace(k, v)
    sys.stderr.write(lg + '\\033[0m\n')


def do_rmdir(dir):
    if '..' in dir:
        die('security')
    if not len(dir.split('/')) > 2:
        die('security')
    if not exists(dir):
        return
    say('removing directory', dir)
    shutil.rmtree(dir)


def do_clear_dir(dir, keep=['.hg*']):
    """clear a workdir
    """
    keep = cast_list(keep)
    # mode 1: all but .hg*
    if '..' in dir:
        die('security')
    dir = os.path.abspath(dir)
    if not os.path.exists(dir):
        dbg('not present', dir)
        return
    if not len(dir.split('/')) > 2:
        die('security')
    for f in os.listdir(dir):
        fn = join(dir, f)
        skip = 0
        for k in keep:
            if fnmatch(f, k):
                skip = 1
                break
        if skip:
            dbg('skipping', f)
            continue
        shutil.rmtree(fn) if isdir(fn) and not islink(fn) else os.unlink(fn)


def do_distri():
    """Linux Distribution Info"""
    res = os.popen('distribution').read()
    if res:
        o, k, arch, distro, version = res.split('|')[:5]
    version = version.split(' ')[0].strip()
    return {
        'os': o,
        'kernel': k,
        'arch': arch,
        'distro': distro,
        'ver': version,
    }


def chmodx(fn):
    st = os.stat(fn)
    os.chmod(fn, st.st_mode | stat.S_IEXEC)


class masked_dir:
    """overlays a directory with an empty temp dir
    useful for e.g. prefix build where we can not rm -rf before, e.g. /opt/nginx
    """

    def __init__(self, d_tgt, d_src=None, exit_remove=True, **kw):
        self.kw = kw
        self.d_tgt = d_tgt
        self.d_src = d_src
        self.exit_remove = exit_remove

    def __enter__(self):
        if not self.d_src:
            self.d_src = tempfile.mkdtemp()
        add_indent('Masking', self.d_tgt, 'with', self.d_src)
        mount(self.d_src, self.d_tgt)
        return self

    def __exit__(self, *a):
        pop_indent('')
        run('umount "%s"' % self.d_tgt)
        if self.exit_remove:
            shutil.rmtree(self.d_src)


def mount(s, t):
    'bind mount. creates mountpoint.'
    s, t = os.path.abspath(s), os.path.abspath(t)
    if not exists(s):
        die('mount fails, source dir not found', s)
    if not exists(t):
        run('mkdir -p "%s"' % t)
    if os.stat(s).st_ino == os.stat(t).st_ino:
        dbg('source already mounted to target', s, '->', t)
        return
    run('mount -o bind "%s" "%s"' % (s, t))


def do_swapon(gigs=2, no_fail=None):
    """creates a /swapfile and activates it.
    If you set gigs to 0 we switch swapping off and delete the file.
    https://www.digitalocean.com/community/tutorials/how-to-add-swap-on-ubuntu-14-04
    """
    fn_s = '/swapfile'
    if not gigs:
        run('swapoff -a; rm -f ' + fn_s)
        return
    if gigs > 4:
        die('max 4 gig. not', gigs)
    if os.path.exists(fn_s):
        run('swapon -s')
        if no_fail:
            return
        die('have already', fn_s)
    # and proceeding with chmod, mkswap & swapon commands. Bingo ! It worked.
    # alternative (fallocate sometimes fail):
    # sudo dd if=/dev/zero of=/myswap count=4096 bs=1MiB
    run(
        [
            'fallocate -l %sG %s' % (gigs, fn_s),
            'chmod 600 ' + fn_s,
            'mkswap ' + fn_s,
            'swapon ' + fn_s,
            'swapon -s',
        ]
    )
    os.system('free -h | grep -v Mem:')
    return True


# ----------------------------------------------------------------------- Tools
def do_deep_copy(src, dest):
    """
    Copy a directory via a tarpipe, i.e. including all perms
    """
    s, d = src, dest
    if not exists(s):
        die('not exists:', s)
    do_mkdir(d)
    run('(cd "%s" && tar -cf - .) | (cd "%s" && tar -xpf -)' % (s, d))


def do_mkdir(directory):
    """ make a directory tree, assuring root directory is present
    """
    d = directory
    if d.startswith('/'):
        # plausi:
        root = '/' + d.split('/')[1]
        if not os.path.exists(root):
            die('Security:', 'Top level dir must exist, does not:', root)
    if not os.path.exists(d):
        run('mkdir -p "%s"' % d)
    return d


# -------------------------------------------- file change monitor
_stats_by_dir = {}


class ChangeDetected(Exception):
    pass


def do_run_on_change(
    monitored_dir,
    depth=1,
    monitored_types='.py',
    check_every=1,
    cmd='echo "change!"',
):
    """
    runs the given command if sth changes in the watched monitored_dir.
    Example: Regenerate help on change.
    ```
    axc os.run_on_change "/xc/bin/axc_app" depth=2 '/xc/bin/axc gen_help'
    ```
    """
    monitored_types = monitored_types.split(',')
    d = os.path.abspath(monitored_dir)
    if not exists(monitored_dir):
        die('not exists', d)
    kw = dict(locals())
    kw['init'] = 1
    check_dir(**kw)
    kw.pop('init')
    do_run = 1
    while 1:
        if do_run:
            if isinstance(cmd, str):
                run(cmd)
            else:
                call(cmd)
        do_run = 0
        sleep(check_every)
        try:
            check_dir(**kw)
        except ChangeDetected as ex:
            say('Change in', ex)
            do_run = 1


def check_files(d, monitored_types, init):
    for f in ls(d):
        if not os.path.isfile(j(d, f)):
            continue
        do = 0
        for t in monitored_types:
            if f.endswith(t):
                do = 1
                break
        if not do:
            continue
        f = j(d, f)
        print(('checking', f))
        old = _stats_by_dir.get(f)
        new = _stats_by_dir[f] = os.stat(f)[8]
        if old != new and not init:
            raise ChangeDetected(f)


def check_dir(monitored_dir, monitored_types, depth, init=0, hir=0, **kw):
    if hir >= depth:
        return
    d = monitored_dir
    old = _stats_by_dir.get(d)
    now = _stats_by_dir[d] = os.stat(d)[8]
    if old != now:
        check_files(d, monitored_types, init)
    for ds in os.listdir(d):
        ds = j(d, ds)
        if not os.path.isdir(ds):
            continue
        check_dir(ds, monitored_types, depth, init, hir + 1)


# ---------------------------------------------------------- OS system commands
def run(
    cmd,
    d='',
    unshared=0,
    no_fail=0,
    fail_msg='',
    dimm=True,
    if_exists=0,
    chrooted=0,
    mounts=0,
    umounts=0,
    quiet=0,
    read=0,
    cd=None,
):
    """ mounts: we mount before the command, permanently.
    Within the command does not work over chroot when rprivate!
    umounts: independent list of clean up umounts after the command
    cd: set cwd parameter to this
    no_fail can be a number then we continue of exit code matches

    """
    if not cmd:
        dbg('no command to run')
        return
    oumounts = umounts

    if chrooted:
        om = mounts
        if mounts in (1, 2):
            # shortcut, one could list those explictly as well::
            mounts = (chrooted, XC, '/proc', '/sys', '/dev', ('/', '/host'))
            if om == 2:
                umounts = mounts
    if if_exists and not exists(os.path.join(d, cmd)):
        dbg('ignoring (not found)', cmd)
        return

    if isinstance(cmd, list):
        cmd = ' && '.join(cmd)
    if d:
        d = 'cd "%s" && ' % d
        cmd = d + cmd

    if chrooted:
        # /bin/bash in debian, in rh seems to be /usr/bin/bash, with still a
        # symlink to /bin though:
        cmd = 'chroot "%s" /bin/bash -c "%s"' % (chrooted, cmd)
    if mounts:
        ums = list(mounts)
        # the chroot as seen from outside:
        mount_prefix = ums.pop(0) or ''
        for um in ums:
            if isinstance(um, str):
                do_mount(um, mount_prefix + um)
            else:
                do_mount(um[0], mount_prefix + um[1])
    if unshared:
        cmd = "unshare -m bash -c '%s'" % cmd
    if read and quiet is 0:  # quiet is default, i.e. not set
        quiet = True
    if quiet < 1:
        dbg('dir:', cd or os.getcwd(), 'cmd:', cmd)
    # cmd = '/usr/bin/env %s' % cmd $ but builtins like export wont work then
    res = 1
    try:

        if read:
            return do_popen(cmd, timeout=read, cd=cd)
        elif quiet:
            res = sp.call(cmd, shell=True, cwd=cd)
        else:
            # colorized run
            fg = 236 if dimm else 249
            # FIXME: we really must do piped command chains,
            # not just && them together (sh # injection)
            ccmd = (
                'res=0;echo -n "\\033[1;38;5;%sm" && ' % fg
                + cmd
                + ' || res=$?; echo -n "\\033[0m"; exit $res'
            )
            # ccmd = 'export term=xterm-256 && ' + cmd
            res = sp.call(ccmd, shell=True, cwd=cd)

    except KeyboardInterrupt as ex:
        die('Keyboard Interrupt. Bye.', no_tb_msg=True)
    finally:
        if umounts:
            ums = list(umounts)
            mount_prefix = ums.pop(0) or ''
            for um in ums:
                if not isinstance(um, str):
                    um = um[1]
                run('umount "%s%s"' % (mount_prefix, um))
    if res == 0:
        return 'ok'
    if res:
        nf = str(no_fail)
        if (nf.isdigit() and res == no_fail) or nf == 'True':
            dbg('exit code', str(res), '- continuing.')
            return
        if fail_msg:
            fail_msg = '[%s]' % fail_msg
        die('failed (exit code %s):' % res, cmd, fail_msg)


def do_popen(cmd, timeout=120, cd=None):
    t1 = time()
    r = {'cmd': cmd}
    r['out'], r['err'] = piped_popen(cmd, cwd=cd).communicate()
    for k in 'out', 'err':
        r[k] = str8(r[k]).strip()
    return r


def rm(rmd, d=None, recreate=0):
    if isinstance(rmd, str):
        rmd = (rmd,)
    for r in rmd:
        if recreate:
            run('find . -mindepth 1 -delete', d=r)
        else:
            run('/bin/rm -rf "%s"' % r, d=d)


def do_kill(match='xxx', sigkill=None):
    """kill process matching 'match'
    if sigkill is set we kill with -9
    if not we check for max 5 seconds if all is gone otherwise we kill
    with sigkill.
    """

    def pids(quiet=0):
        procs = sh.grep(sh.ps('wwwax'), '-i', match)
        if not quiet:
            dbg(procs)
        res = [int(li.strip().split(' ', 1)[0]) for li in procs]
        res = [i for i in res if not i in (base_pid, os.getpid())]
        if quiet:
            dbg(res)
        return res

    sig = -15
    if sigkill:
        sig = -9
    for p in pids():
        try:
            sh.kill(sig, p)
        except Exception as ex:
            warn(p, str(ex))
    t1 = time()
    while time() - t1 < 5:
        if not pids(quiet=1):
            return
        sleep(0.5)
    ps = pids(quiet=1)
    warn('sending SIGKILL', ps)
    for p in ps:
        sh.kill(-9, p)
    if pids():
        die('could not kill', pids())


mount_log = []


def do_mount(src, tgt):
    """
    bind mount a dir to a target
    - creates the target if not exists.
    - checks of already mounted
    """
    s, t = src, tgt
    s, t = os.path.abspath(s), os.path.abspath(t)
    if not exists(s):
        die('mount fails, source dir not found', s)
    if not exists(t):
        if os.path.isdir(src):
            run('mkdir -p "%s"' % t)
        else:
            run('mkdir -p "%s"' % dirname(t))
            run('touch "%s"' % t)
    assert os.path.isdir(s) == os.path.isdir(
        t
    ), 'src and target must be same type, %s %s' % (src, tgt)
    if os.stat(s).st_ino == os.stat(t).st_ino:
        dbg('source already mounted to target', s, '->', t)
        return
    run('mount -o bind "%s" "%s"' % (s, t))
    mount_log.append([s, t])


class mounted(object):
    def __init__(self, src, tgt, indent=True):
        self.src, self.tgt, self.indent = src, tgt, indent

    def __enter__(self):
        if self.indent:
            add_indent('Mounting', self.src, '->', self.tgt)
        return do_mount(src=self.src, tgt=self.tgt)

    def __exit__(self, type, value, traceback):
        run('umount "%s"' % self.tgt)
        if self.indent:
            pop_indent()


def do_unp(
    files,
    output='$PWD',
    dump_command=False,
    forced_unpacker=None,
    list_unpackers=False,
):
    """
    Universal unarchiver (bash alias: 'unp').

    Shamelessly stolen from Armin's unp, stripped of click dep which we find
    overkill for non windows
    """

    class Click:
        class UsageError(Exception):
            pass

        class BadParameter(Exception):
            pass

        def format_filename(self, fn):
            return fn

        def echo(self, s, fg='green', err=False):
            fg = 'red' if err == False else fg
            from axc_app.cli import I, M, R, L

            f = {'green': M, 'red': R, 'yellow': I}.get(fg, L)
            say(f(s))

        secho = echo

    class Ctx(dict):
        def exit(self):
            sys.exit()

    from axc_app.tools import unp, cast_list
    from axc_app.cli import ctx as axc_ctx

    silent = axc_ctx.get('cli_q')
    files = cast_list(files)
    unp.click = click = Click()
    ctx = Ctx()

    if list_unpackers:
        unp.list_unpackers(ctx, 1, 1)

    from axc_app.tools.unp import get_unpacker_class

    unpackers = []

    for filename in files:
        filename = os.path.realpath(filename)
        if not os.path.isfile(filename):
            die('Could not find file "%s".' % click.format_filename(filename))
        if forced_unpacker is not None:
            unpacker_cls = forced_unpacker
        else:
            unpacker_cls = get_unpacker_class(filename)
        unpackers.append(unpacker_cls(filename, silent=silent))

    for unpacker in unpackers:
        if dump_command:
            unpacker.dump_command(output)
        else:
            unpacker.unpack(output)
'''



"""
Just a way to have clean import sequences at app module inits
"""
import os

from gevent import monkey

if not os.environ.get('gevent_no_patch'):
    monkey.patch_all()

fn_backd = lambda: os.environ['var_dir'] + '/proc_enter.sock'


def start_pywsgi_server(application, bind=None, with_backdoor=None):
    """Normal way to start a gevent wsgi server """
    import gevent
    from devapp.app import FLG, app
    from devapp.tools import host_port
    from structlogging import adapters

    # (would be far simpler in threaded bottle, but its a good blueprint
    # for some async http)
    host, port = host_port(bind or FLG.bind)
    s = gevent.pywsgi.WSGIServer
    # devapp adapter to get structlogging from pywsgi plain logging
    server = adapters.adapt_gevent_pywsgi(s((host, port)), logger=app.log)
    server.application = application
    # default logging is structlog, i.e. **kw style:
    fnb = fn_backd()
    if os.path.exists(fnb):
        os.remove(fnb)

    if with_backdoor:
        from gevent.backdoor import BackdoorServer
        from gevent import socket

        listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        listener.bind(fn_backd())
        listener.listen(1)
        # has a ref to the app module if its a devapp:

        bdserver = BackdoorServer(
            listener,
            banner='You are connected into %s' % app.name,
            locals={'app': app.mod},
        )
        gevent.spawn(bdserver.serve_forever)

    app.log.info('Start serving', host=host, port=port)
    server.serve_forever()



import json
import os
import signal
import subprocess
import sys
import time
from functools import partial

from absl import app as abslapp
from absl import flags as flags_
from devapp import gevent_patched  # noqa
from devapp import load, tools
from devapp.lib import sh
from structlogging import sl
from theming.absl_color_help import call_doc, exit_at_help_flag
from theming.colorhilite import coljhighlight

FLG = flags_.FLAGS


env = os.environ
py_env = load.py_env

kvmsg = lambda kw: '  ' + '\n  '.join(['%s: %s' % (k, str(v)) for k, v in kw.items()])
kvprint = lambda l, msg, kw: print('[%s] %s\n%s' % (l, msg, kvmsg(kw)))

notifier = [None]


def notify(app, msg, **kw):
    app.info('NOTIF: %s' % msg, **kw)
    if notifier[0] == None:
        notifier[0] = False
        for n in 'dunstify', 'notify-send':
            if os.system('type %s 1>/dev/null 2>/dev/null' % n) == 0:
                notifier[0] = n
                break
    if not notifier[0]:
        return
    try:
        subprocess.Popen([notifier[0], msg, '\n' + kvmsg(kw)])
    except:
        pass


class App:
    """
    Placeholder for dirs and log and die function, when app is run without the framework
    e.g. in unit tests, where, in the same process later the run_app function will be
    called, for the tested code (e.g lc's test_node_red)

    Otherwise, better use init_app, when you never call run_app.
    """

    is_initted = False
    name = sys.argv[0]

    def die(self, msg, **kw):
        App().warn(msg, **kw)
        sys.exit(1)

    def info(self, msg, **kw):
        kvprint('NFO', msg, kw)

    def warn(self, msg, **kw):
        kvprint('WRN', msg, kw)

    def debug(self, msg, **kw):
        kvprint('DBG ', msg, kw)

    def error(self, msg, **kw):
        kvprint('ERR ', msg, kw)

    def notify(self, msg, **kw):
        notify(App, msg, **kw)

    def __repr__(self):
        return json.dumps({'DevApp': self.name})


App.dbg = App.debug

# def kvs(f):
#    d = [k for k in dir(self) if not k.startswith('_')]
#    return dict([(k, getattr(self, k)) for k in d if f(k)])

# return json.dumps(
#    {
#        'DevApp': self.name,
#        'dirs': kvs(lambda k: '_dir' in k),
#        'attrs': kvs(lambda k: not '_dir' in k),
#    },
#    indent=4,
#    default=str,
#    sort_keys=True,
# )


# this is only for being able to print it:
app = App()


def set_direct_log_methods(app):
    for k in dir(app.log):
        if k[0] != '_' and not k == 'log':
            setattr(app, k, getattr(app.log, k))
    app.dbg = app.debug
    app.log_level = app.log._logger.level


# # allowing from devapp import FLG, flag;  flag.def_str(..)
# class flag:
#     to_flags = tools.define_flags


# r = lambda k: k.replace('DEFINE_', '')
# [setattr(flag, r(k), getattr(flags, k)) for k in dir(flags) if 'DEFINE_' in k]


# --------------------------------------------------------------- Wrapping Apps
def set_dirs():
    """
    - sets PATH if project.root()/bin exists so that resources are found
    """
    from devapp.tools import project

    # no-fail returns None when not found
    db = project.root(no_fail=True)
    if not db:
        return
    db += '/bin'

    if db and os.path.exists(db):
        p = os.environ['PATH']
        if not db + ':' in p:
            os.environ['PATH'] = db + ':' + p
    # e = env.get
    # p = 'DA_DIR_'
    # d = dict(
    #    [
    #        ('%s_dir' % k.split(p, 1)[1].lower(), e(k) + '/' + app_name)
    #        for k in env
    #        if k.startswith(p)
    #    ]
    # )
    ## an app (like build may allow to set --da_dir, overruling env:
    # d['da_dir'] = getattr(FLG, 'da_dir', env.get('DA_DIR', '.'))
    # we are used e.g. at construct/prepare.sh w/o DA:
    # c = 'CONDA_PREFIX'
    # breakpoint()
    # d['pkgs_dir'] = env.get(c + '_1') or env.get(c) + '/pkgs'
    # return d


command_name = lambda: sys.argv[0].rsplit('/', 1)[-1].replace('.py', '')


def set_app(name, log):
    app.log = log
    set_direct_log_methods(app)  # app.info, app.debug, ...
    app.notify = partial(notify, app)
    app.sh = sh
    # app.var_dir ...:
    set_dirs()
    # [setattr(app, k, v) for k, v in dirs(name).items()]
    # allows raise app.die(msg, **kw):
    app.die = app_die(app)
    app.name = name
    app.name_clean = tools.clean_env_key(name)
    app.is_initted = True
    if load.py_env:
        app.env = load.py_env
    if load.app_mod:
        app.mod = load.app_mod

    if FLG.redir_stderr and not os.environ.get('stderr_is_redirected'):
        # we leave it to the the system, can do better than python:
        os.environ['stderr_is_redirected'] = str(FLG.redir_stderr)
        cmd = ' '.join(['"%s"' % j for j in sys.argv])
        sys.exit(os.system(cmd + ' 2>%s' % FLG.redir_stderr))


def init_app_parse_flags(*args):
    """
    An app creator for situations when we do not call run_app - i.e. in pytest
    but still want e.g. logging

    Used in test.tools.build_flow for pytest.
    """
    if hasattr(app, 'log'):
        # not twice
        return
    name = args[0]
    # initializes flags - from argv; stores parsed flags into this FlagValues object.
    # we have no sys.argv in test situations, all will be default:
    l = list(args)
    # allows this: log_level=40 pytest -xs .
    l.extend(['--environ_flags'])
    # FLG = flags.FLAGS
    FLG(l)  # <----------- Flag parsing
    # now reset their values from env
    tools.set_flag_vals_from_env()
    kw_log = {}
    sl.setup_logging(**kw_log)
    log = sl.get_logger(name)
    set_app(name, log)


running = [0]


def run_app(
    main,
    kw_log=None,
    flags_parser=None,
    flags_validator=None,
    wrapper=None,
    flags=None,
    argv=None,
    call_main_when_already_running=False,
):
    """Starter for devapps' main functions

    Examples:
        run = lambda: run_app(build, {'log_dev_fmt_coljson': ['r']})

    flags: Optional flags class

    """

    if running[0]:
        # workaround:
        # certain tests require starting the app when already an app is running:
        if call_main_when_already_running:
            return main()
        app.die('Repeated call of app.run_app', hint='Should call only once per proc.')

    running[0] = True

    # some callers, like plugins, pass over their flags classes, do not call define in mod
    if flags:
        tools.define_flags(flags)
    tools.rm_absl_flags()
    # that's for flags: they only understand helpfull, we want -hf <match>
    argv = argv if argv else sys.argv
    exit_at_help_flag(main, argv)

    #     if '-hf' in argv or '--hf' in argv:
    #         av = list(argv)
    #         argv.clear()
    #         [argv.append('--helpfull' if a in ['-hf', '--hf'] else a) for a in av]

    # name of app (-> logging, var, log, ...folders):
    # n = env.get('DA_CLS') or command_name()
    n = command_name()
    if not flags_parser:
        #     # setup_colorized_help(main, argv)
        flags_parser = abslapp.parse_flags_with_usage
    define_action_flags_in_cli()
    # flags_parser = wrap_flag_parser_with_action_detector(flags_parser)

    try:
        abslapp.run(
            partial(
                run_phase_2,
                name=n,
                main=main,
                kw_log=kw_log,
                flags_validator=flags_validator,
                wrapper=wrapper,
            ),
            argv=argv,
            flags_parser=flags_parser,
        )
    except Exception as ex:
        sys.exit(1)


def define_action_flags_in_cli():
    args = sys.argv
    afs = tools.action_flags
    p = 0
    have = set()
    for a in args[1:]:
        p += 1
        if a == '--':
            return
        if not a in afs:
            continue
        if p > 1:
            pa = args[p - 1]
            pa = pa[2:] if pa.startswith('--') else pa[1:] if pa.startswith('-') else pa
            pf = FLG.__flags.get(pa)
            if pf and pf.flag_type() != 'bool':
                continue
        f = afs.get(a)
        key = f['key']
        if key in have:
            continue
        have.add(key)
        args[p] = '--' + f['key']
        tools.define_flags(f['flg_cls'], sub=key, parent_autoshort=f['autoshort'])
        action[0] = key


action = [0]
# def on_flag_parse_err_have_action_flag(key, af, parser):
#     p = sys.argv.index(key)
#     key = af['key']  # short to long
#     action[0] = key
#     args = list(sys.argv)
#     args[p] = '--' + key
#     tools.define_flags(af['flg_cls'], sub=key, parent_autoshort=af['autoshort'])
#     for arg in args[p + 1 :]:
#         p += 1
#         if arg.startswith('--'):
#             if arg == '--':
#                 break
#             args[p] = arg.replace('--', '--%s_' % key)
#     return parser(args)


#     for k in sys.argv[1:]:
#         if k == '--':
#             break
#         af = afg(k)
#         if k[0] != '-' and af:
#             return on_flag_parse_err_have_action_flag(k, af, parser=parser)
#     # No AF. Let crash - or leave alone, e.g. a app -a1 -- foo -a1 construct
#     sys.stderr.write(err)
#     return sys.exit(1)


class DieNow(Exception):
    # required since sys.exit will be catched - halting the app, not stopping it
    ''


class dev_app_exc_handler(abslapp.ExceptionHandler):
    wants = lambda self, exc: True

    def handle(self, exc):
        if type(exc) == DieNow:
            # app.die was called, we logged already:
            return  # -> silent exit, we logged already
        # Trying to return after json logging as below
        # seems to halt the app with geven sometimes:
        #    48 Sep 10 18:34:55 qtwesacs01 Expert[29848]:   File "src/gevent/greenlet.py", line 716, in gevent._greenlet.Greenlet.run
        #        (...)
        #    Sep 10 18:34:55 qtwesacs01 Expert[29848]:   File "/opt/axwifi_prod/envs/wi
        #    65 Sep 10 18:34:55 qtwesacs01 Expert[29848]:     item: T1 = heapq.heappop(sel
        #    66 Sep 10 18:34:55 qtwesacs01 Expert[29848]: IndexError: index out of range
        #    67 Sep 10 18:34:55 qtwesacs01 Expert[29848]: 2019-09-10T16:34:55Z <Greenlet "
        # so better play save and not invoke app.log when a greenlet crashes:
        raise exc
        # sentry here
        if app.is_initted:
            app.log.error('Exception. Dying now', exc=exc)
        else:
            raise exc


abslapp.install_exception_handler(dev_app_exc_handler())


class Reloaded(Exception):
    pass


def reload_handler(signum, frame):
    raise Reloaded('signal')


from io import StringIO


# def wrap_flag_parser_with_action_detector(flags_parser):
#     def parser(args, p=flags_parser):
#         try:
#             breakpoint()  # FIXME BREAKPOINT
#             e = sys.stderr
#             sys.stderr = StringIO()
#             r = p(args)
#             sys.stderr = e
#             return r
#         except SystemExit as ex:
#             sys.stderr, e = e, sys.stderr.getvalue()
#             return on_flag_parse_err_try_action_flags(args, err=e, parser=p)

#     return parser


# def on_flag_parse_err_try_action_flags(args, err, parser):
#     afg = tools.action_flags.get

#     for k in sys.argv[1:]:
#         if k == '--':
#             break
#         af = afg(k)
#         if k[0] != '-' and af:
#             return on_flag_parse_err_have_action_flag(k, af, parser=parser)
#     # No AF. Let crash - or leave alone, e.g. a app -a1 -- foo -a1 construct
#     sys.stderr.write(err)
#     return sys.exit(1)


# def on_flag_parse_err_have_action_flag(key, af, parser):
#     p = sys.argv.index(key)
#     key = af['key']  # short to long
#     action[0] = key
#     args = list(sys.argv)
#     args[p] = '--' + key
#     tools.define_flags(af['flg_cls'], sub=key, parent_autoshort=af['autoshort'])
#     for arg in args[p + 1 :]:
#         p += 1
#         if arg.startswith('--'):
#             if arg == '--':
#                 break
#             args[p] = arg.replace('--', '--%s_' % key)
#     return parser(args)


# action = ['']


def run_phase_2(args, name, main, kw_log, flags_validator, wrapper):

    tools.set_flag_vals_from_env()  # 0.0001sec

    if FLG.help_call:
        # -h shows level 1:
        call_doc(main, level=2, render=True)
        return

    if FLG.help_call_detailed:
        call_doc(main, level=3, render=True)
        return

    kw_log = {} if kw_log == None else kw_log
    if flags_validator:
        try:
            err = flags_validator()
        except Exception as ex:
            err = str(ex)
        if err:
            print('Flags validation error: %s' % err, file=sys.stderr)
            sys.exit(1)
    sl.setup_logging(**kw_log)
    log = sl.get_logger(name)
    set_app(name, log)
    watcher_pid = None
    if FLG.dirwatch:
        # test if tools present:
        pass

        d, match, rec = (FLG.dirwatch + '::').split(':')[:3]
        d = os.path.abspath(d)
        if not os.path.isdir(d):
            app.die('No directory:', d=d, nfo='Use <dir>:<match>:[r]')
        w = os.path.dirname(os.path.abspath(__file__)) + '/utils/watch_dog.py'
        cmd = [w, ':'.join([d, str(os.getpid()), match, rec])]
        # print(' '.join(cmd))
        watcher_pid = subprocess.Popen(cmd).pid

    # An Action class - use it:
    post = None
    if isinstance(main, type):
        action_cls = main
        if not action[0]:
            for af in tools.action_flags.values():
                if af['flg_cls'].d:
                    action[0] = af['key']
                    app.debug('Running default action', action=af['key'])
                    break
        main = getattr(action_cls, action[0], None)
        if not main:
            app.die('Require action')
        pre = getattr(action_cls, '_pre', 0)
        if pre:
            # app.debug('Prepare hook')
            pre()
            # app.debug(action[0])
        post = getattr(action_cls, '_post', 0)

    res = None
    while True:
        try:
            # so that everybody knows what is running. informational
            app._app_func = main
            # main = lambda: run_app(Action, flags=Flags, wrapper=cleanup)
            res = wrapper(main) if wrapper else main()
            if FLG.dirwatch:
                app.info('Keep running, dirwatch is set')
                signal.signal(1, reload_handler)
                while 1:
                    time.sleep(10)
        except Reloaded as ex:
            continue
        except SystemExit as ex:
            return ex.args[0]
        except KeyboardInterrupt:
            if watcher_pid:
                os.kill(watcher_pid, 9)
            print('Keyboard Interrupt - Bye.')
            sys.exit(1)
            return
        finally:
            if post:
                # app.debug(
                #     'Post app hook', running=[p.__name__ for p in tools.to_list(post)]
                # )

                [p() for p in post] if isinstance(post, list) else post()
        break
    if not isinstance(res, (list, dict, tuple)):
        if not res is None:
            print(res)
        return res
    # exit phase. postprocessing, pretty to stdout, plain to | jq .:
    if FLG.flat:
        res = tools.flatten(res, sep='.')
    if not sys.stdout.isatty():
        jres = json.dumps(res, default=str)
        print(jres)
    else:
        # return res
        print(coljhighlight(res))
    # abseil would print it again:
    # return res


class Die(Exception):
    log = None

    def __init__(self, msg, log=None, **kw):
        log = log or self.log
        if log:
            log.error(msg, **kw)
        else:
            print(
                msg, json.dumps(kw, sort_keys=True, default=str, indent=4)[1:-1],
            )
        raise DieNow()


def app_func(inner=False):
    if not inner:
        return app._app_func
    f = app._app_func
    while hasattr(f, 'func'):
        f = f.func
    return f


def do(func, *a, _step=[0], titelize='', log_level=None, ll=None, fl=None, **kw):
    """When fl (full log level) is set to e.g. 10 we log only the message at higher levels"""

    if ll is not None:
        log_level = {10: 'debug', 20: 'info', 30: 'warn'}.get(ll, 'info')
    fn = func.__qualname__  # .rsplit('.', 1)[-1]
    if titelize:
        _step[0] += 1
        fn = 'STEP %s: %s' % (_step[0], fn)
    log = app.debug if func == system else app.info
    if log_level:
        log = getattr(app, log_level)

    if func == system:
        cmd, args = (a[0] + ' ').split(' ', 1)
        kwl = {} if not args else {'args': args}
        log('sh: ' + cmd, **kwl)
    else:
        if fl is not None and fl < app.log_level:
            log(fn)
        else:
            if a:
                ar = a[0] if len(a) == 1 else a
                log(fn, args=ar, **kw)
            else:
                log(fn, **kw)
    return func(*a, **kw)


def system(cmd, no_fail=False):
    # print('\x1b[38;5;240m', end='')
    d = ' 1>&2'
    # cat -> colors off
    fnf = '/tmp/failed_system_cmd'
    rcmd = 'echo -ne "\x1b[38;5;240m"%s; %s%s || touch "%s"; echo -ne "\x1b[0m"%s'
    rcmd = rcmd % (d, cmd, d, fnf, d)
    os.system(rcmd)
    err = False
    if os.path.exists(fnf):
        os.unlink(fnf)
        err = True
    # print('\x1b[0m', end='')
    if err:
        f = app.warn if no_fail else app.die
        f('Failed', cmd=cmd)
        return 1
    else:
        return 0


# is set into app as .die:
# allows raise app.die(msg, **kw) with correct error logging:
# we want to raise for --pdb_post_mortem
app_die = lambda app: type('Die', (Die,), {'log': app.log})



"""
Using local imports we offer via run function access to functions within
a nested class tree
"""


class API:
    from devapp.components.gitlab import GitLab
    from devapp.components.cloudfoundry import CloudFoundry


def run():
    from devapp.cli import run_cls_tree

    return run_cls_tree(API)


run()



#!/usr/bin/env python
"""

# Spec Building

## Dev
### Frequent abbrs:
- rt: time (e.g. systemd, docker, k8s,)
- cls: A class in a spec python file, denoting a config entity
- fid: Full id of an item (path in the spec : ['Wifi', 'NB01', 'Redis']
"""
# stream of specs - from anywhere, we will  as daemon some day:
# before that its just a one item stream, with item pushed at proc stasrt:


# fmt:off
import json
import os
import shutil
import sys
import uuid
from functools import partial

import rx
from devapp.app import FLG, app
from devapp.spec import fs_components as fs
from devapp.spec import os_tools, tools
from devapp.spec.templ import SmartAdapt, align_by_eq, exec_templ, r_objs_by_fid
from devapp.tools import (
    abspath,
    camel_to_snake,
    dirname,
    exec_file,
    exists,
    into,
    read_file,
    to_list,
    write_file,
)
from devapp.utils.py_source_env import source
from devapp.utils.rx_tools import Ops, chain, do_if, map_diff_log
from rx import operators as op
from rx.operators import filter, flat_map
from tree_builder import full_id, struct_props
from tree_builder.links import conns_by_cls_dependent
from tree_builder.links import map_connections_to_all_classes as map_connections

# fmt:on


here = abspath(dirname(__file__))

FS = tools.FS

# Our default mapping op should log diffs:
Ops.default = map_diff_log

env = os.environ

environ = os.environ
# flag.string('spec_file', '', 'Spec file')
#
# flag.string('cls_match', '$DA_CLS', 'Restrict operation to matching classes')
# flag.string('log_time_fmt', 'dt', 'Time format of logs. E.g. "ISO", "%m:%s"')
# flag.string('dir', '$DA_DIR', 'Solution directory')
# flag.string(
#    'from_dir',
#    '',
#    'Source directory, witch conf/SPEC.py.\nWill be copied over to dir before install.',
#    short_name='f',
# )
#

# fmt:off
fidstr       = lambda r: '.'.join(r['spec']['fid'])  # fullid (path) as str
pfidstr      = lambda r: fidstr(r).rsplit('.', 1)[0] # parent full id
props        = lambda r: r['spec']['props']
is_svc       = lambda r: props(r)['type'] == 'Service'
is_app       = lambda r: props(r)['type'] == 'App'
is_exe       = lambda r: is_svc(r) or is_app(r)
dir_setup    = lambda r: app.var_dir + '/' + fidstr(r)
da_dir       = lambda r: r['env']['DA_DIR']          # DevApps Dir (=CondaPrefix)
runtime      = lambda r: props(r).get('runtime', 'systemd' if is_exe(r) else None)
slfn         = lambda fn: ('/' + fn) if fn else ''
build_dir    = lambda r: r['env']['build_dir']
comp_dir     = lambda r, fn='': build_dir(r) + '/fs_components' + slfn(fn)
unit_name    = lambda r: '%(name)s-%(DA_PREFIX)s.service' % r['env']
registry_url = env.get('DA_URL_REGISTRY','')
# fmt:on


def load_spec():
    g = exec_file(app.da_dir + '/conf/SPEC.py', get_src=True)
    return {'root': g['root'], 'src': g['source']}


# if not set we install the whole project - into a container:
def ldd(r):
    breakpoint()
    return r


# we push the 'r' dicts, built from the spec a few times through, mutating
# This is the list we'll use within rx.from_:
r_objs = []


def classes_stream(spec):
    def create_classes(obs, _, spec=spec, cls=None):
        """from root scan thru, pump out"""
        cls = spec['root'] if cls is None else cls
        fid = full_id(cls)
        fids = fidstr({'spec': {'fid': fid}})
        app.warn(cls.name, fid=fids, type=cls.type)
        props = dict(struct_props(cls))
        props['filesystem'] = cls.filesystem = to_list(props.get('filesystem'))
        for k in props['filesystem']:
            # can be given as string, then we find the rest
            if isinstance(k, str):
                fs.fs_comps[k] = {'name': k}  # will be filled by fs
                # strings only we DO normalize to lists:
            else:
                n = k.get('name')
                # leave like it was in spec:
                fs.fs_comps[n] = dict(k)

        r = {
            'cls': cls,
            '___': fids,  # it should show up on top when we pp r objs
            'env': {},  # will go to env.sh, sourced by systemd
            'spec': {'props': props, 'fid': fid},
            'links': {},  # links.json
            'fs': {},  # fs.json
        }
        r_objs.append(r)
        r_objs_by_fid[fids] = r
        obs.on_next(r)
        [create_classes(obs, _, spec, c) for c in cls._childs]
        if cls._root == cls:
            obs.on_completed()

    return rx.create(create_classes)


def add_init_env(r, da_env):
    r['env'].update(dict(da_env))
    r['env']['DA_CLS'] = r['___']
    return r


def run_rs_filtered(msg, *obs, cls_match=''):
    """
    filter: prefix match
    """

    def matched(o, cls_match=cls_match):
        if not cls_match:
            return o
        return o if o['___'].startswith(cls_match) else None

    matched = lambda o: o
    if cls_match:
        if not any([o for o in r_objs if matched(o)]):
            app.die('No spec classes match', match=cls_match, mode='startswith')
    app.info(msg)
    p = rx.from_(r_objs).pipe(filter(matched))
    return chain(p, *obs)


# ----------------------------------------------------------------------- links


def build_connections(r):
    """
    Our job: Add 'links', as pure data, telling the
    template exactly what to do
    That data is provided by interfaces and meta information of the repos
    interfacing with each other...
    """
    r['links'] = conns_by_cls_dependent(r['cls'])
    return r


def add_base_dir_vars(r):
    fid, e = fidstr(r), r['env']  # AXWiFi.Dispatcher
    for d in fs.base_dirs:
        bd = da_dir(r) + '/%s' % d  # /opt/axwifi/log
        # /opt/axwifi/log/AXWiFi.Dispatcher:
        e[d + '_dir'] = '%s/%s' % (bd, fid)
    SD = e.get('DA_DIR_STATS')
    e['stats_dir'] = e['var_dir'] if not SD else '%s/%s' % (SD, fid)
    e['pidfile'] = r['env']['var_dir'] + '/pidfile'
    return r


def trace_if(match):
    def t(r, match=match):
        print(r.get('___'))
        breakpoint() if match.lower() in str(r).lower() else 0
        return r

    return t


def find_cfg_dir(r):
    """where are the templates and all that"""
    rfs = r.get('fs_comps')
    if not rfs:
        # no rfs
        return r
    rfs = rfs[-1]  # last one, its a stack and last one must have config
    # the last one is it:
    r['env']['run_dir'] = rfs['run_dir']
    # remember
    dco = r['env']['DA_DIR_CHECKOUT'] = rfs['checkout_dir']
    # trace_if('nginx')(r)
    if exists(dco + '/recipe'):
        r['dir_cfg_checkout'] = dco + '/recipe'
    sd = 'daemon' if is_svc(r) else 'exe' if is_exe(r) else ''
    if not sd:
        return r
    # n = '' when /daemon or /exe contains the config
    # todo: have to see more variants to find that dir
    tries = (props(r)['name'], props(r).get('base'), rfs['base_name'], '')
    for nt in tries:
        if not nt:
            continue
        nt = camel_to_snake(nt)
        h = []
        for n in nt, nt.replace('_', ''):
            if n in h:
                continue
            h.append(n)
            ns = ('/' + n) if n else ''
            cd = '%s/%s%s' % (dco, sd, ns)
            if sd == 'daemon':
                cd += 'd'
            if exists(cd):
                r['dir_cfg_checkout'] = cd
                r['env']['cfg_path'] = pth = cd[len(dco) + 1 :]
                r['env']['daemon_dir'] = r['env']['run_dir'] + '/' + pth
                r['daemon_name'] = n
                return r
    return r


def _try_add_single_conffile(e, r_cfg):
    d = r_cfg

    # set conffile automatically if there is only one pointint to conf_dir:
    if exists(d + '/templates'):
        for k in ('conf', 'etc'):
            cd = k + '_dir__'
            l = [
                c.split(cd, 1)[1]
                for c in os.listdir(d + '/templates')
                if c.startswith(cd)
            ]
            # no subdir, we take exactly one file in conf_dir, no axc subdir:
            # (hashi understands conf_dir and parses all .json, so we put into axc)
            l = [c for c in l if not '__' in c]
            if len(l) == 1 and 'conffile' not in e:
                e['conffile'] = e[k + '_dir'] + '/' + l[0]


def req_unshare(r):
    # systemd seems to not export the mounts :-)
    return False


def add_app_env(r):
    """ man systemd.directives """
    e, da_dir = r['env'], env['DA_DIR']
    s = 'service_'
    if not e.get('app'):
        return r
    e[s + 'working_directory'] = wd = da_dir + '/run'

    _ = 'app_run_script'
    r['env'][_] = rnr = '.'.join(reversed(r['___'].lower().split('.')))
    e['app_run_exe_link'] = '%s/%s' % (wd, e['DA_CLS'])
    e['functions'] = e.get('functions', {})
    if not 'start' in e['functions']:
        e['functions']['start'] = [e['app']] + e.get('app_args', [])

    se = s + 'exec_start'
    if not se in e:
        app.info('Adding exec start directive', se='start')
        e[se] = 'start'
    e['app_run_script_path'] = da_dir + '/build/bin/' + rnr
    return r


def_paths = lambda r: (
    # apps can render templates into this;
    r['env']['etc_dir'] + '/bin',
    # E.g. the 'app' tool
    r['env']['DA_DIR_DEVAPPS'] + '/bin',
    # Suprocesses (.e.g bash) or poll success handler should have easy access to the starters:
    r['env']['DA_DIR'] + '/build/bin',
    '/usr/bin',
    '/bin',
    '/usr/sbin',
    '/sbin',
)

def_py_paths = lambda r, v=sys.version_info: (
    r['env']['DA_DIR_DEVAPPS'] + '/python',
    r['env']['DA_DIR_ENV_DEVAPPS']
    + '/lib/python%s.%s/site-packages' % (v.major, v.minor),
)


def set_env(r):
    """adding repo props (via its env.sh), then spec props to final env.sh"""
    e = r['env']
    ce = _is_not_in_conda_base_env(r)
    if ce:
        # good to know, e.g. in tempaltes(nginx)
        e['DA_DIR_ENV'] = ce

    # setting DA_DATACENTER, DA_NODE ...:
    # cls = r['cls']
    # for p in cls._parents[1:]:
    #    e['DA_%s' % p.type.upper()] = fidstr(p)
    r_cfg = r.get('dir_cfg_checkout')
    if r_cfg:
        _try_add_single_conffile(e, r_cfg)
        fn = r_cfg + '/env.sh'
        env_sh = read_file(fn, '', strip_comments=True)
        if env_sh:
            app.debug('reading', fn=fn)
            if '%(' in env_sh:
                try:
                    env_sh = env_sh % SmartAdapt(r, fn)
                except Exception as ex:
                    if not sys.stdin.isatty():
                        raise
                    print(ex)
                    print('breakpoint set at env render')
                    breakpoint()
                    keep_ctx = True
            # real sourcing, i.e. executing functions, interpreting vars
            try:
                fn = '/tmp/devapp_build.%s.%s' % (r['___'], uuid.uuid4())
                fn = write_file(fn, env_sh)
                e.update(os_tools.source(fn))
            except SystemExit as ex:
                msg = ''
                if ' =' in env_sh or '= ' in env_sh:
                    msg = 'Note that key=val assignments in bash '
                    msg += 'must be done w/o spaces'
                    msg += ' (have found " = ").'
                app.die('Could not shell source', fn=fn, msg=msg)
            finally:
                os.unlink(fn)

    # add the paths which the fs components loader found, per component:
    # for $PATH: systemd does not replace $PATH - i.e. we must add ours at
    # runtime.:
    rfs = fs.rfs_comps(r)
    for k in 'PATH', 'PYTHONPATH':
        P = []
        # all components, e.g. repos
        for comp in rfs:
            P.extend(comp.get('env_' + k, []))
        if P:
            have = e.get(k, '')
            if have:
                # hard set? that rules:
                P.insert(0, have)
        # TODO: We must do all 4, suse has different /bin then /usr/bin
        # better way is to set e.g. add_PATH and add only at runtime:
        if k == 'PATH':
            for p in def_paths(r):
                P.append(p) if not p in P else 0
            for p in solution_libs('bin', r):
                P.insert(0, p)
        elif k == 'PYTHONPATH':
            for p in def_py_paths(r):
                P.append(p) if not p in P else 0
            for p in solution_libs('python', r):
                P.insert(0, p)
        if P:
            e[k] = ':'.join(P)

    # TODO: Add repo meta.py props of all filesystem(?)
    e.update(r['spec']['props'])
    rt = props(r).get('runtime')
    if not rt:
        return r
    elif rt == 'systemd:boot':
        # todo: normalize this with other rts, this seems a bit too hardcoded:
        e['app'] = 'systemd-nspawn'
        l = ['--register=false', '-D', e['build_dir'] + '/fs', '--boot']
        e['app_args'] = l
    e['name_unit'] = unit_name(r)
    return r


def solution_libs(typ, r, have={'_scanned_da_dir_lib': 0}):
    if have['_scanned_da_dir_lib']:
        return have.get(typ) or ()
    d = env['DA_DIR'] + '/lib'
    if os.path.exists(d):
        for l in os.listdir(d):
            have[l] = [d + '/' + l]
    have['_scanned_da_dir_lib'] = 1
    return solution_libs(typ, r)


def ld(r):
    return r


def get_devapps_env_vars():
    # those will end in env.sh and are parametrizing our build - nothing else:
    daenv = dict([(k, env[k]) for k in env if k.startswith('DA_')])
    d = daenv['DA_DIR']
    # prefix is the directory name we run in, not the full path.
    # if people want two version of foo then they must say foo.lab, foo.14.1:
    daenv['DA_PREFIX'] = daenv.get('DA_PREFIX', d.rsplit('/')[-1])
    return daenv


def render_templates(r):
    e = r['env']
    d = r.get('dir_cfg_checkout')
    if not d:
        return r
    for td, ends in (d, '.tmpl'), (d + '/templates', ''):
        if not exists(td):
            continue
        for fn in os.listdir(td):
            if (
                not '__' in fn  # dir sep. always there
                or fn.startswith('.')
                or not fn.endswith(ends)
                or fn.endswith('.bak')
                or '__pycache__' in fn
            ):
                app.debug('Ignoring', fn=fn)
                continue
            res = exec_templ(td + '/' + fn, r)
            fnr = fn
            for k in 'etc', 'conf':
                fnr = fnr.replace(k + '_dir__', e[k + '_dir'] + '/')
            if 'conda_env_dir' in fnr:
                cd = [
                    c['checkout_dir'] for c in r['fs_comps'] if c['type'] == 'conda_env'
                ]
                if not cd:
                    raise Exception('No conda env', fnr)
                fnr = fnr.replace('conda_env_dir', cd[0])
            fnr = fnr.replace('__', '/')
            fnt = comp_dir(r, fn)
            write_file(fnt, res, mkdir=1)
            r['fs'][fnr] = {'from': fnt, 'meth': 'symlink'}

    return r


# def _add_base_site_packages_to_pypath(env):
#    """Base utilities of the base / devapps conda python setup should be there
#    always and are py3.x compat"""
#    sv = '%s.%s' % (sys.version_info.major, sys.version_info.minor)
#    bp = env['DA_DIR'] + '/lib/python%s/site-packages' % sv
#    # systemd unit starts -m devapp.run -> devapp nededs to be seen:
#    md = abspath(env['DA_DIR_DEVAPPS'] + '/../python')
#    for pp, p in ('sys_path_append', bp), ('PYTHONPATH', md):
#        env[pp] = ':'.join((env[pp], p)) if env.get(pp) else p
#


def _add_path(env, conda_env):
    """Binaries (non devapp modules) are childs of the python starter - need to
    be found in run.py"""
    p = conda_env + '/bin'
    env['PATH'] = p + ':' + env['PATH'] if env.get('PATH') else p


# def write_env(r):
#    e = r['env']
#    conda_env = _is_not_in_conda_base_env(r)
#    if conda_env:
#        # _add_base_site_packages_to_pypath(e)
#        _add_path(e, conda_env)
#    # DIR again, its already in run but easy access via env sourcing:
#    # if one is sourceing THIS he wants it exported, e.g. a unit:
#    s = ['#!', 'set -a', 'DA_DIR="%(DA_DIR)s"' % e]
#    ses = 'service_exec_start'
#    if is_exe(r) and not ses in e and not props(r).get('runtime') == 'user':
#        raise Exception('No $service_exec_start defined')
#
#    for k, v in sorted(e.items()):
#        # we source this to get in, if initted would be there then we could not
#        if k in ('DA_DIR', 'DA_INITTED'):
#            continue
#        v = (
#            "'%s'" % json.dumps(v)
#            if isinstance(v, (list, dict))
#            else '"%s"' % v
#        )
#        # we support in spec to say e.g.
#        # app_args = ['-c', '%(etc_dir)s/nginx.conf']
#        if '%(' in v:
#            v = v % e
#        kv = '%s=%s' % (k, v)
#        s += [kv]
#    s += ['set +a', '']
#    fn = e['build_dir'] + '/env.sh'
#    write_file(fn, s, mkdir=True)
#    return r
#

# fmt: off
env_defaults = {
    'unit_stop_when_unneeded'  : 'false',
    'service_type'             : 'simple',
    'service_kill_signal'      : 'SIGTERM',
    'service_send_SIGKILL'     : 'yes',
    'service_private_tmp'      : 'true',
    'service_timeout_stop_sec' : '5',
}
# fmt: on
def add_unit_vars(r):
    """these are added after env.sh written -> only make it into rendering ctx"""
    e = r['env']
    e['unit_description'] = (
        e.get('unit_description') or '%(type)s %(name)s [%(DA_CLS)s]' % e
    )
    [into(e, k, e.get(k, v)) for k, v in env_defaults.items()]
    e['service_PID_file'] = e['var_dir'] + '/pidfile'
    # not unit_name, since that would end in the unitfile:
    return r


def render_unit(r):
    res = exec_templ(here + '/templates/unit', r)
    res = align_by_eq(res.splitlines())
    fn = units_dir().replace('/', '__') + '__' + r['env']['name_unit']
    fnt = comp_dir(r, fn)
    write_file(fnt, res, mkdir=1)
    r['unit_file'] = fnt
    return r


units_dir = (
    lambda: '/etc/systemd/system'
    if not os.geteuid()
    else '%(HOME)s/.config/systemd/user' % env
)


def link_unit_in_parent_or_host(r):
    """Where to write the unit file to. We go back all parents - if one has
    a runtime which boots then we add our unit file there - else on the host
    Remember that parents can be isolation abstractions only - w/o actual rts
    """
    d_systemd = units_dir() + '/'

    pfid, pr = r['___'], None
    while len(pfid.split('.')) > 1:
        pfid = pfid.rsplit('.', 1)[0]
        _ = r_objs_by_fid[pfid]
        if props(_).get('runtime'):
            pr = _
            break
    # if no parent has a runtime we create the host link now:
    run_fn = d_systemd + r['env']['name_unit']
    r['env']['unit_file'] = run_fn
    if not pr:
        # no parent - we write unit into host:
        if not exists(d_systemd):
            err = os.system('mkdir -p "%s"' % d_systemd)
            if err:
                app.die('Could not create', d=d_systemd)

        # too many levels of symbolic links -> copy:
        app.info('Copying', frm=r['unit_file'], to=run_fn)
        FS.copy_file(r['unit_file'], run_fn)
    else:
        # have parent - we write unit into build dir of parent:
        rt = props(pr)['runtime']
        if ':boot' in rt:
            # write our link into the parent, so when it builds its fs before start
            # we are were we should be:
            # (grand..)parent runtime boots
            # fs.root_fs_mount_ph: '$fs', placeholder for root fs mount point
            root = fs.root_fs_mount_ph
            pr['fs'][root + run_fn] = {'from': r['unit_file'], 'meth': 'mount'}
            for target in ('devapps', 'multi-user'):
                pr['fs'][
                    root
                    + '%s%s.target.wants/' % (d_systemd, target)
                    + r['env']['name_unit']
                ] = {'from': run_fn, 'meth': 'symlink'}
        else:
            app.die('Not supported nesting into', runtime=rt)
    return r


fn_app_run_script = '%(DA_DIR)s/build/bin/%(app_run_script)s'


def add_app_run_script_to_build_dir(r):
    """
    DA_DIR/build/bin/ scripts.
    This should be in $PATH, for the user not for systemd
    Single commands, runnable from even outside conda:
    TODO: we could replace this thing with a 2 liner function per service as well
    """
    s = '''#!%(DA_DIR)s/run/%(DA_CLS)s
    # Offers to be run from the host w/o conda, w/o env.sh sourced, like in
    # systemd-unit => We source the env in python:
    import sys
    sys.path.append("%(DA_DIR_DEVAPPS)s/python")
    from devapp import load
    load.app_load(fn_env="%(DA_DIR)s/build/%(DA_CLS)s/env.json")
    from devapp import run
    run.app_run(args=sys.argv)
    '''
    s = s.replace('\n    ', '\n')
    fn = fn_app_run_script % r['env']
    write_file(fn, s % r['env'], mkdir=True)
    # os.chmod(fn, stat.S_IEXEC)
    os.chmod(fn, 0o755)
    return r


_ = [
    add_unit_vars,
    render_unit,
    link_unit_in_parent_or_host,
    add_app_run_script_to_build_dir,
]
render_runtime_starters = {'systemd': _, 'systemd:boot': _}


def _write_file(r, name, param):
    bd = build_dir(r)
    if not exists(bd):
        os.mkdir(bd)
    s = json.dumps(r[param], indent=4, default=str)
    # _cfg = ['-c', '%(etc_dir)s/runner.toml']
    if '%(' in s:
        # this will fail with any % in it -> then need to double replace:
        # or insert by string splits or re sub
        s = s % r['env']
    write_file(bd + '/' + name, s)
    return r


# fmt:off
write_env = lambda r: _write_file(r, 'env.json', 'env')
write_links = lambda r: _write_file(r, 'links.json', 'links')
write_fs_stack = lambda r: _write_file(r, 'fs_stack.json', 'fs')
# fmt:on


def create_var_dir(r):
    app.sh.mkdir('-p', r['env']['var_dir'])
    return r


def execute_persistent_fs_changes(r):
    fs = r.get('fs', {})
    if not fs:
        return r
    FS.build_from_fs_stack(mode='shared', app_env=r['env'], fs=fs)
    return r


def _is_not_in_conda_base_env(r):
    fsc = r.get('fs_comps', ())
    envs = [e for e in fsc if e.get('type') == 'conda_env']
    if len(envs) > 1:
        app.die('Cannot have 2 conda envs at same time', envs=envs)
    if envs:
        return envs[0]['run_dir']


def named_link_to_python(r):
    """This is for the app starters' shebang. We take the python from devapps env if no
    python in service env"""
    p = _is_not_in_conda_base_env(r) + '/bin/python'
    if not exists(p):
        p = env['DA_DIR_ENV_DEVAPPS'] + '/bin/python'
    FS.symlink(p, r['env']['app_run_exe_link'])
    return r


def clean_old_build(env):
    shutil.rmtree(env['DA_DIR'] + '/build/bin', ignore_errors=True)


def source_da_dir_env():
    D = env['DA_DIR']
    # env.sh shall understand any DA_ var in the build environ:
    d = source(D + '/env.sh', ign_missing=True, insert_env_vars=('DA_*',))
    env.update(d)


def source_devapps_env_if_not_activated(dir=None):
    """We may call da spec.build from the host, w/o
    properly activating base conda -> missing the sourcing of activate.d/devapps.sh
    -> do that now:
    """
    dir = env['DA_DIR'] = app.da_dir = dir or FLG.da_dir
    if env.get('DA_DIR_REPOS') and env.get('DA_DIR_DEVAPPS'):
        return
    d_base = sys.argv[0].split('/envs/', 1)[0]
    fn = d_base + '/etc/conda/activate.d/devapps.sh'
    if not exists(fn):
        raise Exception(
            'Missing DEVAPPS vars. Properly install devapps into a conda environ and activate it'
        )
    env['PATH'] = d_base + '/bin:%(PATH)s' % env
    env.update(source(fn, insert_env_vars=('PATH',)))


def copy_over_on_from_dir(d):
    if not d:
        return
    d = abspath(d)
    if not exists(d + '/conf/SPEC.py'):
        app.die('Not found', dir=d + 'conf/SPEC.py')
    FS.tarpipe(d, env['DA_DIR'], create=True)


def build(dir=None, from_dir=None, cls_match=None):
    source_devapps_env_if_not_activated(dir)
    copy_over_on_from_dir(from_dir)
    source_da_dir_env()
    devapp_env = get_devapps_env_vars()
    # this creates an iterable of all spec classes, serialized (into 'r' dicts),
    # which is # repeatedly used as a base observable, in run_rs invokations
    # below:
    run_rs = partial(run_rs_filtered, cls_match=cls_match)
    chain(
        rx.just(load_spec()),
        op.do_action(lambda spec: map_connections(spec['root'])),
        flat_map(classes_stream),
    ).run()
    clean_old_build(devapp_env)
    fs.check_mk_dirs(devapp_env)
    if not fs.fs_comps:
        app.warn('Nothing to build, no components. Bye.')
        sys.exit(0)
    fs.fetch_fs_comps(devapp_env).run()
    run_rs(
        'Preparation',
        partial(add_init_env, da_env=devapp_env),
        build_connections,
        add_base_dir_vars,
        do_if(
            # add OUR filesystem components into our data:
            lambda r: into(r, 'fs_comps', fs.rfs_comps(r)),
            find_cfg_dir,
            set_env,
            add_app_env,
            fs.define_fs_stack,
            if_=lambda r: props(r).get('filesystem'),
        ),
    ).run()
    run_rs(
        'Configuration',
        render_templates,
        do_if(render_runtime_starters, which=lambda r: runtime(r)),
    ).run()
    run_rs(
        'Writing files',
        write_fs_stack,
        write_links,
        execute_persistent_fs_changes,
        write_env,
        create_var_dir,
        do_if(named_link_to_python, if_=lambda r: r['env'].get('app_run_exe_link'),),
    ).run()
    # only outside possible:
    if os.environ.get('CONTAINER_NAME'):
        mode = 'container'
    else:
        if os.geteuid():
            mode = 'systemd(user mode)'
            app.sh.systemctl('--user', 'daemon-reload')
        else:
            mode = 'systemd(system mode)'
            app.sh.systemctl('daemon-reload')
    from tree_builder import UserInput

    UserInput.store()
    app.info('Spec Built', mode=mode)
    # run_rs('Pull Filesystem Resources', group_by(props('fil


# def set_cli_flags():
#    # user or config.cls did not set this environ var, then we assume we
#    # are root:
#    if FLG.cls_match == '$DA_CLS':
#        FLG.cls_match = ''
#    env['DA_DIR'] = FLG.dir
#    return
#
#
# main = lambda: run_app(
#    build,
#    {'log_dev_fmt_coljson': ['r', 'rdiff']},
#    flags_validator=set_cli_flags,
# )
#
if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Common API for all resources

"""
# FIXME Do this as state table! Its crying for that

import os, sys
import json
import time
from functools import partial
import importlib
from devapp.app import app, do, run_app, system
from devapp.tools import (
    offset_port,
    to_list as listed,
    FLG,
    project,
    deindent,
    exists,
    repl_dollar_var_with_env_val,
    dirname,
    write_file,
    read_file,
)

T_unit = '''
[Unit]
Description      = %(descr)s %(name)s
Wants            = network-online.target
StopWhenUnneeded = false

[Service]
Type             = simple

ExecStart        = %(exec_start)s
KillSignal       = SIGTERM
PrivateTmp       = true
Restart          = always
RestartSec       = 5
SendSIGKILL      = yes
TimeoutStopSec   = 5
SyslogIdentifier = %(name)s

[Install]
WantedBy = default.target

# _MATCH_ (auto created %(ctime)s) 
'''
unit_match = 'DevApp Unit'
T_unit = T_unit.replace('_MATCH_', unit_match)


class S:
    """state"""

    rscs_defined = None  # from resource_list
    pkg_cmd = None
    rsc_modified = False
    name_to_func = {}
    conda_prefix = None  # env vars replaced
    fs_dir = None  # filesystem install dest
    constants = {}
    rsc_dirs = {}


def add_const(key, val, skip_exists=True):
    """import order matters, i.e. 4A's flows file will overrule lc-py"""
    v = S.constants.get(key)
    if v is not None:
        app.warn('Overwriting constant %s' % key, have=v, new=val)
    S.constants[key] = val


def constant(key, dflt=None):
    v = S.constants.get(key, dflt)
    if v is None:
        app.die('Expected constant not found', key=key)
    return v


def into(m, k, v):
    m[k] = v


environ = os.environ


def cur_prefix():
    cli = os.environ.get('CONDA_PREFIX', '').split('/envs/', 1)[0]
    if not cli:
        app.die('No $CONDA_PREFIX currently set')
    return cli


def set_conda_prefix():
    """
    Resources install location, except filesystem based ones. Env vars resolved.

    Aliases:
    - local|l: <project_dir>/conda
    - default|d: $HOME/miniconda3 (default path of conda)
    - current|c: Any current conda_prefix set when running.
    
    Note: Installing resources outside the project keeps the project relocatable and resources reusable for other products.
    """

    cli = repl_dollar_var_with_env_val(FLG.conda_prefix)
    if not cli:
        cli = 'default'
    if cli in ('current', 'c'):
        cli = cur_prefix()
    elif cli in ('local', 'l'):
        cli = project.root() + '/conda'
    elif cli in ('default', 'd'):
        cli = os.environ['HOME'] + '/miniconda3'
    cli = os.path.abspath(cli)
    S.conda_prefix = cli


def set_fs_dir():
    """
    Filesystem based resource location. Env vars resolved.
    Aliases:
    - local|l: <project_dir>/fs
    - default|d: $HOME/miniconda3/fs (default path of conda)
    - conda|c: Within conda_prefix/fs
    """
    cli = FLG.fs_dir or 'default'
    cli = repl_dollar_var_with_env_val(cli)
    if cli in ('local', 'l'):
        cli = project.root() + '/fs'
    elif cli in ('default', 'd'):
        cli = S.conda_prefix + '/fs'
    elif cli in ('conda', 'c'):
        cli = cur_prefix() + '/fs'
    cli = os.path.abspath(cli)
    S.fs_dir = cli


conda_prefix = lambda: S.conda_prefix or [set_conda_prefix(), S.conda_prefix][1]


class CommonFlags:
    autoshort = ''

    class install_state:
        n = 'show install state infos'
        d = False

    class conda_prefix:
        n = set_conda_prefix.__doc__
        d = 'default'

    class fs_dir:
        n = set_fs_dir.__doc__
        d = 'default'

    class log_resources_fully:
        n = 'Always output all settings of resources when logging'
        d = False


# --------------------------------------------------------------------- tools/  actions

simple_types = (bool, int, float, str, list)
to_struct = lambda s: (to_list if s.name == 'resources' else to_dict)(s)
to_list = lambda rscs: [to_dict(rsc) for rsc in rscs]
to_dict = lambda rsc: {
    k: getattr(rsc, k)
    for k in dir(rsc)
    if not k.startswith('_') and isinstance(getattr(rsc, k), simple_types)
}


# # get some nice respresentation of the rsc classes:
# class Resource(type):
#     # print the resources always fully - too much spam:
#     # __repr__ = __str__ = lambda s: json.dumps(to_struct(s), sort_keys=True, indent=2)
#     __repr__ = __str__ = (
#         lambda s: rsc_repr(s)
#         if not FLG.log_resources_fully
#         else json.dumps(to_struct(s), sort_keys=True, indent=2)
#     )


# class R(metaclass=Resource):
#     listed = classmethod(lambda s: rsc_classes(s))


# def representable(name, *bases):
#     return type(name, (R,) + bases, {'__metaclass__': Resource})


# api:
g = lambda rsc, key, default='': getattr(rsc, key, default)


def gf(rsc, key, default=''):
    """Get function if it was a callable"""
    v = g(rsc, key, default)
    return S.name_to_func.get('.'.join((rsc.name, v)), v)


# ------------------------------------ Setting Directories project dir and conda_prefix


# def matching_resource(match, exact=False):
#     if not S.rscs_defined:
#         find_resource_defs()
#     if exact:
#         rsc = [r for r in S.rscs_defined.listed() if match == r.name]
#     else:
#         app.info('trying fuzzy match', match=match)
#         rsc = [r for r in S.rscs_defined.listed() if match in str(to_dict(r)).lower()]
#         app.debug('matches', rsc=rsc)
#         if len(rsc) > 1:
#             prov = to_list(g(rsc, 'provides', ''))
#             rsc = [r for r in rsc if match in r.name.lower() or match in str(prov)]
#             if len(rsc) > 1:
#                 rsc = [r for r in rsc if match == r.name or match in prov]
#     if not rsc or len(rsc) > 1:
#         app.die('No unique match', match=match, defined=S.rscs_defined, found=len(rsc))

#     return rsc[0]


is_fs = lambda rsc: str(g(rsc, 'pkg')).startswith('layers:')


def dir_rsc_cfg(rsc):
    """configured directory of the resource"""
    if is_no_pkg_rsc(rsc):
        return project.root() + '/bin'
    elif is_fs(rsc):
        return S.fs_dir + '/' + rsc.name
    else:
        return S.conda_prefix + '/envs/%s/bin' % (g(rsc, 'conda_env', rsc.name))


def rsc_path(rsc, verify_present=False):
    """
    Find path of resource (e.g. /home/joe/miniconda3/envs/myrsc/bin)
    Return nothing if not present

    We are intrested in the path not the file itself because we'll export it before running.
    """
    path = dir_rsc_cfg(rsc)
    if not verify_present:
        return path
    v = gf(rsc, 'verify_present')
    if v:
        return path if v(rsc=rsc, path=path) else None
    if is_no_pkg_rsc(rsc):
        return path if exists(path + '/' + rsc.name) else None
    exe = g(rsc, 'exe') or g(rsc, 'cmd')
    if not exe:
        return path if exists(path) else None
    else:
        return path if exists(path + '/' + exe) else None


interactive = lambda: '-y' if FLG.force else ''
# resources without a package (e.g. client)
is_no_pkg_rsc = lambda rsc: g(rsc, 'pkg') == False


class Run:
    def get_full_cmd(rsc, sel):
        cmd_pre = ''
        if is_no_pkg_rsc(rsc):
            f = gf(rsc, sel, sel)
            d = project.root()
        else:
            d = rsc_path(rsc, verify_present=True)
            if not d:
                # d = rsc_path(rsc, verify_present=True)
                app.die('Not installed', **to_dict(rsc))

            spec = {}
            # find a function to call:
            f = gf(rsc, sel, sel)
            if not callable(f):
                f = gf(rsc, 'run')
                if not callable(f):
                    f = gf(rsc, 'cmd')
        # else just take the command string:
        cmd = gf(rsc, 'run') or gf(rsc, sel, sel)
        if callable(f):
            spec = f(cmd=sel, rsc=rsc, pth=d, api=api())
            i = isinstance
            if i(spec, dict):
                # e.g.: export LD_...
                cmd_pre = spec.get('cmd_pre', '')
            cmd = spec if i(spec, str) else (spec or {}).pop('cmd', cmd)
            if callable(cmd):
                # cmd was a function which e.g. just changed dir:
                cmd = cmd.__name__

        # app.info('Adding PATH', path=d)
        # os.environ['PATH'] = d + ':%s' % os.environ.get('PATH')
        if cmd.startswith(':'):
            cmd = cmd[1:]
        else:
            cmd = d + '/' + cmd
        cmd = cmd_pre + cmd
        # if not cmd.split(' ')[0] == orig_cmd.split(' ')[0]:
        #     app.info('completing command', given=orig_cmd, completed=cmd)
        return cmd, spec


class Install:
    """Install methods"""

    _ = 'https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh'
    conda_installer_url = _

    def no_pkg(rsc):
        app.info('No package resource - no install')

    def resource(rsc):
        """Api entry: install a resource"""
        Install.requirements(g(rsc, 'req'))
        if not rsc.installed or g(FLG, 'force_reinstall', ''):
            if g(rsc, 'pkg') == False:
                install_mode = Install.no_pkg
            elif is_fs(rsc):
                install_mode = Install.FS.filesystem
            else:
                install_mode = Install.Conda.conda_env

            do(install_mode, rsc=rsc)
        else:
            app.debug('already installed', rsc=rsc.name)
        do(Install.post, rsc=rsc, ll=10)
        do(Install.write_starter_and_unit_file, rsc=rsc, ll=10)
        rsc.installed = True
        S.rsc_modified = True

    class Tools:
        def download(url, dest):
            # TODO
            cmd = 'curl "%s" -o "%s"' % (url, dest)
            do(system, cmd)
            app.info('downloaded', dest=dest)

    class FS:
        def filesystem(rsc):
            if not rsc_path(rsc, verify_present=True):
                pass
            d = S.fs_dir + '/%s' % rsc.name
            img = rsc.pkg.split('layers:', 1)[1]
            system('ops container_pull --repo "%s" --dir "%s.img"' % (img, d))
            s = '--skip_filesystem_adaptions'
            system('ops container_build --dirs "%s.img" --target_dir "%s" %s' % (d, d, s))

    def write_unit_file(name, fn, rsc):
        pn = project.root().rsplit('/', 1)[-1]
        m = {
            'name': name,
            'descr': '%s %s ' % (g(rsc, 'n', name), pn),
            'ctime': time.ctime(),
            'exec_start': fn,
        }
        d = os.environ['HOME'] + '/.config/systemd/user'
        if not exists(d):
            app.info('creating systemd user dir', dir=d)
            os.makedirs(d)
        n_svc = '%s-%s.service' % (name, pn)
        sfn = d + '/' + n_svc
        unit = T_unit % m
        have = read_file(sfn, dflt='')
        if unit != have:
            write_file(sfn, unit)
            app.info('have written unit file', fn=sfn)
        else:
            app.debug('unit file unchanged', fn=sfn)
        return n_svc

    def write_starter_and_unit_file(rsc):
        env = os.environ
        env['d_conf'] = project.root() + '/conf'
        env['d_bin'] = project.root() + '/bin'

        def write(cmd, fcmd, spec):
            # if cmd == 'dot':
            #    breakpoint()  # FIXME BREAKPOINT
            if isinstance(spec, str):
                spec = {}
            app.debug('writing bin/' + cmd, cmd=fcmd)
            pre_exec = spec.pop('pre_exec', '')
            spec_env = spec.pop('env', {})
            fn = project.root() + '/bin/%s' % cmd
            have = read_file(fn, '')
            marker = '-AUTOCREATED- '
            if have and len(have.split('\n' + marker, 1)[0].splitlines()) > 8:
                return app.warn(
                    'Skipping write of starter file, marker was manually removed',
                    fn=fn,
                    marker=marker,
                )
            call = sys.argv[0]
            args = ' '.join(sys.argv[1:])
            if call.endswith('/ops'):
                call = call.rsplit('/', 1)[-1] + ' ' + args
            else:
                call = call + '\\\n' + args
            r = [
                '#!/usr/bin/env bash',
                '',
                '# Delete line containing "%s" to avoid overwrites of this file at project init'
                % marker,
                "_='%s" % time.ctime(),
                marker,
                '%s' % call,
                "'",
            ]
            add = r.append
            units = g(FLG, 'init_create_unit_files', [])
            if g(FLG, 'init_create_all_units'):
                units.extend(listed(g(rsc, 'systemd', None)))

            has_unit = False
            if any([u for u in units if u == cmd]):
                has_unit = True
                n_svc = Install.write_unit_file(cmd, fn, rsc)
                s = '''

                case "${1:-}" in
                    start|restart|stop|status)
                        systemctl --user --no-pager $1 "%s"
                        exit $?
                        ;;
                esac
                '''
                add(deindent(s % n_svc))

            # env['PATH'] = '%s:$PATH' % g(rsc, 'path')
            add('')
            add("H='__HOME__'")
            add('export PROJECT_ROOT="%s"' % project.root())
            add('')
            add('# Resource settings:')
            # for whoever needs that:
            allk = set()
            for m in to_dict(rsc), spec, spec_env:
                for k, v in m.items():
                    allk.add(k)
                    if k == 'port':
                        v = offset_port(v)
                    exp = 'export ' if m == env else ''
                    add('%s%s="%s"' % (exp, k, str(v)))
            add('')
            if 'logdir' not in allk:
                add('export logdir="$PROJECT_ROOT/log"')
            if 'logfile' not in allk:
                add('export logfile="$logdir/%s.log"' % cmd)

            add('')
            # only for services. Tools not: (e.g. cd mydir && git status -> bum, differentdir)
            if has_unit:
                add('cd "%s"' % project.root())
            # for client but also possibly other python resources:
            env['PYTHONPATH'] = env.get('PYTHONPATH', '')
            add('export PYTHONPATH="%(d_conf)s:%(PYTHONPATH)s"' % env)
            # to just call w/o virtenv activation:
            # we remove all d_bin refs and put them to the end, to avoid circulars,
            # e.g. for npm - find it first in conda env
            d, p = env['d_bin'], env['PATH']
            for k in ':', '':
                p = p.replace(d + k, '')
            p += ':' + d
            add('export PATH="$path:%s"' % p)
            add('')

            add('return 2>/dev/null # when sourced\n')

            if pre_exec:
                for c in listed(pre_exec):
                    add(c)
                add('')
            postproc = g(FLG, 'add_post_process_cmd', '')
            if not has_unit:
                postproc = ''
            add('%s "$@" %s\n\n' % (fcmd, postproc))

            h = os.environ['HOME']
            s = '\n'.join(r).replace(h, '$H').replace('__HOME__', h)

            if have and have.split(marker, 1)[-1] == s.split(marker, 1)[-1]:
                app.debug('No change', cmd=cmd)
            else:
                write_file(fn, s, chmod=0o755)

        for cmd in rsc_cmds(rsc):
            fcmd, spec = Run.get_full_cmd(rsc, cmd)
            cmd = cmd.rsplit('/', 1)[-1]  # /usr/bin/seagull -> bin/seagull
            write(cmd, fcmd, spec)

    def requirements(req):
        if not req:
            return
        app.info('Installing requirements', req=req)
        for r in to_list(req):
            rsc = matching_resource(r, exact=True)
            Install.resource(rsc)

    class Conda:
        def conda_env(rsc):
            D = S.conda_prefix

            # return False if not dp else all([exists(dp + '/' + p) for p in rsc.provides])
            if not exists(D + '/bin/'):
                do(Install.Conda.base, location=D)

            if str(rsc.path).startswith(D):
                return app.debug('already installed - skipping', rsc=rsc)

            env = g(rsc, 'conda_env', rsc.name)
            ctx = dict(D=D, name=env, yes=interactive())
            # source is not working for $0 = sh:
            cmd = [
                '. "%(D)s/etc/profile.d/conda.sh"',
                'conda create %(yes)s -n "%(name)s"',
                'conda activate "%(name)s"',
            ]
            pth = '%(D)s/envs/%(name)s/bin/' % ctx

            if g(rsc, 'typ') == 'pip':
                ctx['cmd'] = rsc.cmd
                cmd += ['conda install python; %p/pip install %%(cmd)s' % pth]
            else:
                icmd = g(rsc, 'conda_inst', '')
                if icmd:
                    cmd += [icmd]
                else:
                    p = g(rsc, 'conda_pkg') or g(rsc, 'pkg') or ' '.join(rsc.provides)
                    chan = g(rsc, 'conda_chan', '')
                    if chan:
                        chan = '-c ' + chan
                    ctx['chan'] = chan
                    ctx['pkg'] = p
                    cmd += ['conda install %(yes)s %(chan)s %(pkg)s']
            cmd = ' && '.join(cmd) % ctx
            rsc.path = g(rsc, 'path') or pth
            return do(system, cmd)

        def base(location):
            app.warn('conda dir not found', dir=location)
            if not FLG.force:
                q = 'Confirm: Install miniconda at %s? [y/N]'
                if not input(q % location).lower() == 'y':
                    app.die('unconfirmed')
            fn = os.environ['HOME'] + '/install_miniconda.sh'
            url = Install.conda_installer_url
            if not exists(fn):
                Install.Tools.download(url, fn)
            if not exists(fn):
                app.die('download failed', fn=fn)
            os.system('chmod +x "%s"' % fn)
            os.makedirs(os.path.dirname(location), exist_ok=True)
            do(system, '%s -b -p "%s"' % (fn, location))

    def post(rsc):
        pf = gf(rsc, 'post_inst')
        if not pf:
            return
        # when post_inst_verify is True they have been installed. When not present there are none:
        if g(rsc, 'post_inst_verify') and not g(FLG, 'force_reinstall'):
            return

        if callable(pf):
            app.info('postinstall function', name=rsc.name)
            # they may be requirements for the post install as well:
            Install.requirements(gf(rsc, 'post_inst_req'))
            # run it:
            pth, here = environ['PATH'], os.getcwd()
            try:
                if rsc.path.endswith('/bin'):
                    environ['PATH'] = rsc.path + ':' + pth
                res = pf(rsc, install=True, api=api())
            finally:
                os.chdir(here)
                environ['PATH'] = pth

            # Usually pf installs - except:
            # special case for tools, it has not a single cmd but a list of provides.
            # no cmd -> not yet install_conda_env was run. we do it now via this conventin:
            if res == 'install_conda_env':
                return Install.Conda.conda_env(rsc)
        else:
            app.die('Unsupported postinstall method', method=pf)


def rsc_cmds(rsc):
    r = []
    cmd = g(rsc, 'exe') or g(rsc, 'cmd')
    if cmd:
        r.append(cmd)
    for p in g(rsc, 'provides', ()):
        r.append(p)
    return r


def api():
    from devapp import resource_tools as api

    return api


def check_installed(rscs):
    """Sets the installed flag"""

    def check_installed_path(rsc):
        rsc.path = rsc_path(rsc, verify_present=True) or False

    def check_post_installed(rsc):
        # funnily we cannot import this module there, hash(S) is different then :-/
        # guess because of dynamic imports / freezing. So we pass ourselves over:
        cp = gf(rsc, 'post_inst')
        if callable(cp):
            res = cp(rsc, verify=True, api=api())
            rsc.post_inst_verify = res
        elif cp:
            raise NotImplementedError

    for rsc in listed(rscs):
        d = partial(do, rsc=rsc, ll=10)
        # if g(rsc, 'pkg') == False:
        #     rsc.installed = False
        #     continue
        d(check_installed_path)
        d(check_post_installed)

        rsc.installed = bool(rsc.path and g(rsc, 'post_inst_verify', 'y'))


add_install_state = check_installed


def find_resources_files_in_sys_path():
    files = {}

    def find_file_in_pth(pth, d, files=files):
        fn = '%s/%s/operations/resources.py' % (pth, d)
        if exists(fn) and not d in files:
            files[d] = fn
            S.rsc_dirs[fn] = pth + '/' + d

    def find_files_in_sys_path(pth):
        [find_file_in_pth(pth, d) for d in os.listdir(pth)]

    [find_files_in_sys_path(pth=pth) for pth in sys.path if os.path.isdir(pth)]
    return files


def complete_attrs(rsc, fn):
    def to_name(rsc, p):
        if callable(p):
            n = p.__name__
            S.name_to_func['.'.join((rsc.name, n))] = p
            return n
        return p

    def repl_callable(rsc, k, v):
        listed_attrs = ['provides']  # allow to say provides= npm
        if isinstance(v, list):
            vn = [to_name(rsc, i) for i in v]
        else:
            vn = to_name(rsc, v)
        if k in listed_attrs and not isinstance(vn, list):
            vn = [vn]
        if vn != v:
            setattr(rsc, k, vn)

    rsc.doc = rsc.__doc__ or ''
    rsc.name = rsc.__name__
    rsc.module = rsc.__module__.replace('.operations.resources', '')
    rsc.__repr__ = lambda r: str(to_dict(r))
    rsc.__str__ = lambda r: to_str(r)
    rsc.module_dir = S.rsc_dirs[fn]
    rsc.disabled = g(rsc, 'disabled', g(rsc, 'd', False))
    rsc.installed = g(rsc, 'installed', False)
    [repl_callable(rsc, k, getattr(rsc, k)) for k in dir(rsc) if not k.startswith('_')]


def to_str(rsc):
    svc = g(rsc, 'systemd')
    i = 'i' if g(rsc, 'installed') else ' '
    d = 'd' if rsc.disabled else ' '
    s = 's' if svc else ' '
    n = svc if svc else rsc.name
    return '%s %s %s %s %s' % (s, i, d, n, g(rsc, 'provides', ''))


def rsc_repr(rsc):
    return rsc.name + ('[i]' if g(rsc, 'installed', '') else '')


# ------------------------------------------------------------------------    API entry
def find_resource_defs(_have_mod={}):
    """

    Delivers back a list of resources with unique classnames, i.e. w/o parents of same name

    (class mysql of 4A overwrites mysql of python)

    Identical ones are also possible and are removed:

        class rsc(other_rsc):
            pass


    """
    set_conda_prefix()
    set_fs_dir()
    m = {}
    for k in 'fs_dir', 'conda_prefix':
        f, s = g(FLG, k), g(S, k)
        m[k] = s if f == s else '%s(%s)' % (s, f)
    app.info('Directories', **m)

    rsc_files = find_resources_files_in_sys_path()

    def rsc_classes(rsc_cls_of_resources_module):
        """rscs is class rsc (with resource classes as attrs"""
        _ = rsc_cls_of_resources_module
        rsc_clses = [g(_, k) for k in dir(_) if not k.startswith('_')]
        rsc_clses = [r for r in rsc_clses if isinstance(r, type)]
        return rsc_clses

    rscs = []

    def find_cls_in_pth(d, fn, rscs=rscs):
        try:
            n_mod = '%s.operations.resources' % d
            if n_mod in _have_mod:
                app.info('%s' % n_mod, conflict=fn, taken=_have_mod[n_mod])
                return
            _have_mod[n_mod] = fn
            mod = importlib.import_module(n_mod)
        except Exception as ex:
            return app.die('Cannot import', fn=fn, exc=ex)
        mod.rsc._package = d
        mod.rsc._filename = fn

        rsc_clses = rsc_classes(mod.rsc)
        [complete_attrs(r, fn) for r in rsc_clses]

        rscs.extend(rsc_clses)

    {find_cls_in_pth(d, fn) for d, fn in rsc_files.items()}

    def remove_redefined_rscs(rscs=rscs):

        # this removes the identical ones:
        all = set(rscs)
        # assert len(all) == len(rscs)
        for r in all:
            for p in r.mro()[1:]:
                if p in all:
                    n = p.__name__
                    d = {'redefined by': r.module, 'was': p.module + '.' + n}
                    app.info('Resource redefined: %s' % n, **d)
                    rscs.remove(p)

        i, j = len(all), len(rscs)
        if i > j:
            app.info('%s resources redefined' % (i - j))

    remove_redefined_rscs()
    # remove dubs (through class rsc(other_rsc))
    rscs = set(rscs)
    # sort by name:
    rscs = sorted([r for r in rscs], key=lambda x: x.name)
    # now we make singletons for __repr__ w/o metaclass hacks
    rscs = [r() for r in rscs]
    for k in rscs:
        app.debug(k.name, **to_dict(k))
    S.rscs_defined = rscs
    return rscs



#!/usr/bin/env python
import time
from copy import deepcopy
from functools import partial
from threading import current_thread
from typing import Callable, Iterable

import pycond
import rx
from devapp.app import app
from devapp.tools import funcname, in_gevent, jdiff, now
from rx import operators as op
from rx.scheduler import NewThreadScheduler, ThreadPoolScheduler
from theming.colorhilite import coljhighlight as pretty

# Custom ones also return an rx one, they wrap inner items:
is_operator = lambda o: getattr(o, '__module__', '').startswith('rx')

threaded = lambda t=10: op.observe_on(ThreadPoolScheduler(t))

pass_ = lambda *ev, **kw: 0


class Ops:
    """Allowing apps to set the default operator in chains"""

    default = op.map


def do_if(*actions: Callable, if_=None, which=None, default_op=None):
    """
    # Conditional Running of Operators
    Run the actions stream only when (alternatively):
    - the if_ condition is met
    - actions is a *dict* of actions, where the key is evaled by 'which' selector

    ## Params
    - actions: An operator chain, to be run when if_ evals to truthy
               OR
               A dict of operator chains, to be run when the which selector
               evals to dict key

    - if_, which: selector functions
    - default_ops: operator to wrap plain functions into (default rx's map
                   operator)
    Examples:

        We want to map to mult, then divide only on odd numbers:

        rx.from_([2,3,4]).pipe(
            do_some_stuff,
            do_if(mult, divide, if_=lambda i: i % 2)
        )

        We want to map to mult, then divide only on 3:

        rx.from_([2,3,4]).pipe(
               do_some_stuff,
               do_if(
                    {3: (mult, divide)},
                    which=lambda i: i
                 ))
    I.e. on the given stream 2,3,4 those produce the same results.
    The latter example demonstrates that we simply do nothing when the selector
    doese not match

    """

    def dict_actions(key, actions_by_key):
        if callable(actions_by_key):
            acts = actions_by_key(key)
        else:
            acts = actions_by_key.get(key)
        if not acts:
            return ()
        return acts if isinstance(acts, (list, tuple)) else (acts,)

    if if_:
        actions_by_key = {True: actions}
        # convenience for the user: we just have 2 cases, and group_by should not
        # see (and have to cache) anything not falsy:
        if not callable(if_):
            pc = pycond.pycond(if_)
            selector = lambda x: pc(state=x)
        else:
            selector = lambda x: bool(if_(x))
    else:
        if not isinstance(actions[0], dict) and not callable(actions[0]):
            raise Exception(
                (
                    'First argument must be a dict of operators,'
                    'based on the "which" selector result'
                )
            )
        actions_by_key = actions[0]
        selector = which

    return rx.pipe(
        op.group_by(selector),
        op.flat_map(
            lambda s, dop=default_op: chain(
                s, *dict_actions(s.key, actions_by_key), default_op=dop
            )
        ),
    )


def only(if_, pipe: Iterable[Callable]):
    return do_if(*pipe, if_=if_)


def chain(observe, *obs, default_op=None):
    return stream(observe=observe, pipe=list(obs), default_op=default_op)


def stream(observe, pipe, default_op=None):
    p = p1 = observe
    d = Ops.default if default_op is None else default_op
    ops = []

    for o in pipe:
        if is_operator(o):
            ops.append(o)
        else:
            ops.append(d(o))

        # p = p.pipe(o if is_operator(o) else d(o))
        # p = p.pipe(o if is_operator(o) else d(o))
    p = p.pipe(*ops)
    return p


def log(ev):
    app.info('Event', json=ev)
    return ev


def show(ev, add_dt=False):
    m = {'event': ev, 'thread': current_thread().name, 'ts': time.ctime()}
    if add_dt:
        ev['meta']['dt'] = now() - ev['meta']['t0']
    print(pretty(m))
    return ev


def pdb(ev):
    show(ev)
    print('''Breakpoint inserted''')
    breakpoint()
    return ev


def map_diff_log(f, wrap=op.map):
    """a map showing diff before and after
    Requires app set up
    """
    fn = funcname(f)

    def diff_log_map(r, f=f, fn=fn):
        try:
            ro = deepcopy(r)
            res = f(r)
        except Exception as ex:
            app.log.error(fn, r=r, diff=jdiff(ro, r))
            raise
        app.log.info(fn, rdiff=jdiff(ro, r))
        return res

    return wrap(diff_log_map)


def new_thread(c=[]):
    if c:
        return c[0]
    if in_gevent():
        # in compex nested blocking arrangements with high load we spotted
        # rx heap errors when using NewThreadSched. when monkey patched
        # => really use this one, no errs here:
        from rx.scheduler.eventloop import GEventScheduler

        s = GEventScheduler()
    else:
        s = NewThreadScheduler()
    c.append(s)
    return s


def concurrent(pipe):
    """Puts a function call, including a subsequent processing chain onto a
    new thread/greenlet
    The function is the first item of the pipe.
    There is no control over concurrency, this is isolated runs, flat-mapped
    back into the main stream

    # without the func = 0 element of pipe it would be:
    return stream(rx.just(r), [op.observe_on(new_thread()), *pipe])
    """

    def block(r, func, ops):
        s = rx.from_callable(lambda r=r: func(r), scheduler=new_thread())
        return s.pipe(*ops)

    return op.flat_map(partial(block, func=pipe[0], ops=pipe[1:]))


# ----------- StreamRunner (streams from serializable dicts) ------------------
def op_by_dict_cfg(opdict, stream_runner):
    """
    Configure operators with dicts, e.g.:
        def foo(i:int)
    configured like:
        {"foo": {"i": 10}}

    Higher order function operators like
        def foo(i, pipe)
    configured like:
        {"foo": {"pipe": ["bar"], "i": 10}}
    and bar is resolved into the function

    -> The resolution into funcs requires the 'pipe' keyword.
    """
    cls = stream_runner
    op_name, v = func_name_args_from_dict(opdict)

    f = find_operator_by_name(op_name, cls)
    if 'pipe' in v:
        # we don't want to ditch the serializable config,
        # want to keep as sent from the config machinery
        v = dict(v)
        # recursion here:
        v['pipe'] = pipe_from_cfg(stream_runner, v['pipe'])
        return f(**v)
    else:
        return partial(f, **v)


def find_operator_by_name(oprtr, stream_runner):
    cls = stream_runner
    for f in cls.functions:
        if isinstance(oprtr, list):
            return op_by_dict_cfg({'concurrent': {'pipe': oprtr}}, cls)
        if isinstance(oprtr, dict):
            return op_by_dict_cfg(oprtr, cls)

        # We do allow nested functionality classes with a namespace:
        parts = oprtr.split('.')
        op = getattr(f, parts.pop(0), None)
        while parts and op:
            part = parts.pop(0)
            op = getattr(op, part, None)
        if op:
            return op
    raise Exception('operator not found', {'operator': oprtr, 'checked': cls.functions})


def pipe_from_cfg(stream_runner, cfg_pipe):
    """Turns a list of operator names into their functions
    looking them up int the stream_runner._functions
    """
    cfg_pipe = [cfg_pipe] if not isinstance(cfg_pipe, list) else cfg_pipe
    return [find_operator_by_name(op, stream_runner) for op in cfg_pipe]


def func_name_args_from_dict(single_item_dict):
    # {'some_func_in_namespace': {'args': 'foo'}}
    for k, v in single_item_dict.items():
        return k, v


class StreamRunner:
    _subscriptions = {}
    _configs = {}

    @classmethod
    def run_cfg(cls, name, cfg):
        """We store the subscriptions in order to get rid of them when streams
        are redefined"""
        if name in cls._subscriptions:
            app.warning('Stopping', name=name)
            cls._subscriptions[name]['subs'].dispose()
        cls._configs[name] = cfg
        return cls.run(name)

    @classmethod
    def run(cls, name, observe=None, pipe=None, **kw):
        cfg = cls._configs.get(name, {})
        if cfg:
            obs = cfg['observe']
            if isinstance(obs, str):
                observe = getattr(cls, cfg['observe'])
            else:
                # func -> args, e.g. {'from_route': {'path': '/inform', 'type': 'blocking', 'verbs': ['post']},
                func_name, args = func_name_args_from_dict(obs)
                observe = args['subject'] = rx.subjects.Subject()
                # Reference the stream in the Streams class:
                setattr(cls, name, observe)
                # use the same function finding machinery as for the pipe ops:
                op_by_dict_cfg({func_name: args}, cls)()

            pipe = pipe_from_cfg(cls, cfg['pipe'])

        app.warning('Starting stream', name=name, json=cfg)

        stream_def = lambda observe=observe, pipe=pipe: stream(observe, pipe)
        stream_ = stream_def()
        d = stream_.subscribe(pass_)
        cls._subscriptions[name] = {
            'stream': stream_,
            'def': stream_def,
            'subs': d,
            'observe': observe,
            'pipe': pipe,
        }



# FIXME
fixme = '''
Remove all this.
If colors are important, use the xml output of --helpfull.
[2020-10-07 22:16] Done. Still due though:
This has a lot of legacy in trying to colorize functions and nested class trees help output
in sync wiht absl cli flags.
(do_action of old devapps)

Remove the bloat sooner or later...
'''

# TODO: Display use cases and detailled help attributes (see devapp/app.py flags)

import io
import os
import sys
import textwrap
from inspect import signature
from xml.etree import ElementTree as ET

from absl import app, flags
from devapp.tools import tabulate, termwidth, action_flags, define_flags


def func_from_partial(f):
    while not hasattr(f, '__module__'):
        if not hasattr(f, 'func'):
            return f
        f = f.func
    return f, ''


def call_doc(obj, level=2, render=False):
    """ Used by devapps.app.py for -hh and -hhh"""
    raise Exception('call_doc not implemented')
    f, kw = func_from_partial(obj)
    if type(f) == type:
        out = class_doc(f, level=level)
    else:
        out = func_doc(obj, level=level + 1)
    if len(out) < 3:
        breakpoint()  # FIXME BREAKPOINT
        if 'lambda' in out[0]:
            out.pop(0)
        mod = obj.__module__
        fn = sys.modules[mod].__file__
        out.insert(0, 'No flags defined: %s (%s)' % (mod, fn))
    out = '\n'.join(out)
    out = out.strip()
    if render:
        mdvl.render(out, header_numbering=True, header_numb_level_min=2)

    return out


deindent = lambda d: ('\n'.join([l.strip() for l in d.splitlines()]).strip()) + '\n'


def func_doc(obj, level=1):
    f, _ = func_from_partial(obj)
    n = f.__qualname__
    if '.' in n:
        n = n.split('.', 1)[1]
    sig = str(signature(obj))

    # hack: specical case for devapp.run external process:
    # there we want to show the args of the process
    if n in ('run_exit', 'app_run_exit'):
        h = 'run: ' + (' '.join([str(i) for i in obj.args[0]]))
    else:
        h = '%s %s' % (n, str(sig).replace('*', '.'))
    r, doc = [h], f.__doc__ or ''
    if level > 1:
        r.append(deindent(doc))
    return r


def class_doc(cls, level=1, hir=0, out=None):
    hir += 1
    out = [] if out is None else out
    out.append('%s %s' % ('#' * hir, cls.__name__))
    d = getattr(cls, '__doc__')
    if d and level > 1:
        out.append(deindent(d))
    dflt_meth = getattr(cls, 'default', None)
    clss = []
    aliases = []
    funcs = []
    for k in sorted(dir(cls)):
        if k.startswith('_') or k == 'default':
            continue
        c = getattr(cls, k)
        if isinstance(c, type):
            if k != c.__name__ and c.__name__ in dir(cls):
                aliases.append([k, c.__name__])
                continue
            clss.append(c)

        elif callable(c):
            funcs.append(
                {'short': '', 'long': k, 'expl': '\n'.join(func_doc(c, level=level)),}
            )
    # FIXME
    add_md_list(funcs, out)
    for c in clss:
        class_doc(c, level, hir, out)
    if aliases:
        out.append('\nAliases: ' + ', '.join(['%s->%s' % (i, k) for i, k in aliases]))
    out.append('- -hh[h] to get more details about calls.')

    if hir == 1:
        return out


def module_doc(mod):
    n, d = mod.__name__, mod.__doc__
    if n == '__main__':
        n = sys.argv[0].rsplit('/', 1)[-1]
    return '# %s\n%s' % (n, d)


# --------------------------------------------------------------- -hf/--helpfull output
def term_widths(have_match):
    tw = termwidth()
    widths = dict(name=int(tw * 0.2 + 0.5), short_name=6, default=int(tw * 0.15 + 0.5))
    last = 'meaning'
    if have_match:
        widths = dict(name=25, short_name=8, meaning=0)
        last = 'default'

    widths[last] = max(0, tw - sum(widths.values()))
    widths['term'] = tw
    return widths


def term_line(line_spec, widths, have_match):
    """One line of output - have_match typically [main module name] """
    m = line_spec
    is_action = m.get('action')
    # if is_action:
    #     breakpoint()  # FIXME BREAKPOINT
    r = []
    pos = 0
    for colmn in 'short_name', 'name', 'default', 'meaning':
        v = str(m.get(colmn)) or ''
        if v == 'None':
            v = '' if colmn == 'short_name' else "''"

        if have_match:
            v += ' '
        w = widths[colmn]
        oldpos = pos
        pos += w
        if colmn == 'meaning':
            choi = m.get('choices')
            if choi:
                choi = ' [%s]' % ','.join(choi)
                if have_match or len(v) + len(choi) <= w:
                    v += col('choices') + choi
            det = m.get('details')
            if det:
                # det = ' 🗎 ' + det
                det = '\n' + det
                if have_match or len(v) + len(det) <= w:
                    v += col('details') + det
        c = colmn
        if is_action:
            if m.get('action_flag'):
                if colmn in ('short_name', 'meaning'):
                    v = '   ' + v
            elif colmn == 'default':
                v, c = 'ACTION', 'action'
                if m.get('is_default'):
                    v += '*'
                    c = 'default_action'

        if have_match:
            if colmn == 'meaning':
                r[-1] = r[-1].rstrip()
                v = '\n%s' % v  # (' ' * widths['short_name']) + v
            # else:
            #     if 0 and len(v) > w:
            #         dedented_text = textwrap.dedent(v).strip()
            #         v = textwrap.fill(
            #             dedented_text,
            #             initial_indent=' ' * oldpos,
            #             subsequent_indent=' ' * 4,
            #             width=widths['term'],
            #         )
            #         v += ' '
            #         pos = 0
        else:
            # if colmn == 'default' and 'mini' in v:
            #    breakpoint()  # FIXME BREAKPOINT
            if len(v) > w:
                if colmn in ('name', 'short_name'):
                    v = v + '\n' + ' ' * pos
                else:
                    try:
                        v = textwrap.shorten(v, width=w, placeholder='..')
                    except Exception as ex:
                        # when not even placeholder fits
                        v = v[:w]
        r.append(col(c) + v.ljust(w))
    return ''.join(r)


def to_terminal(flags, widths, match=None):
    r = [term_line(line_spec, widths, have_match=match) for line_spec in flags]
    return '\n'.join(r)


def parse_xml_help(xml_help, match, cli_actions=None):
    """parse --helpxml output of absl"""

    class types:
        def do_float(c, el):
            c['default'] = float(c['default'])

        def do_bool(c, el):
            c['default'] = False if c['default'] == 'false' else True

        def do_string(c, el):
            m = c['meaning']
            c['meaning'] = m  #'\x1b[1;38;5;124m%(meaning)s\x1b[0;m' % c

        def do_comma_separated_list_of_strings(c, el):
            pass

        def do_string_enum(c, el):
            c['choices'] = [e.text for e in el]
            c['meaning'] = c['meaning'].split('>: ', 1)[-1]

        def do_multi_string_enum(c, el):
            return types.do_string_enum(c, el)

        def do_int(c, el):
            c['default'] = int(c['default'])

    tree = ET.fromstring(xml_help)
    m, items = {}, {}
    r = {'program': tree[0].text, 'usage': tree[0].text, 'flags': m}
    for fs in tree[2:]:
        f = fs
        c = {}
        for t in f:
            tn = t.tag
            c[tn] = t.text
            if tn == 'type':
                c[tn] = c[tn].replace(' ', '_')
                if not c[tn] == 'multi_string':
                    getattr(types, 'do_' + c[tn])(c, f)
                mn = c['meaning']
                # Details already set by the flag instantiation in define_flags:
                if 'Details: ' in mn:
                    c['meaning'], c['details'] = mn.split('Details: ', 1)
                break
        if match and not any([m in str(c) for m in match]):
            continue
        c.pop('current', 0)
        l = m.setdefault(c['file'], [])
        l.append(c)
        items[c['name']] = {'pos': len(l) - 1, 'mod': c['file']}

    have = set()
    for f, af in action_flags.items():
        k = af['key']
        if k in have:
            continue
        have.add(k)
        item = items.get(k)
        p, mod = item['pos'] + 1, item['mod']
        m[mod][p - 1]['action'] = True
        m[mod][p - 1]['is_default'] = af['is_default']
        # afp.append([p - 1, mod, k, True])
        if af in cli_actions:
            f = m[mod][p]
            while f['name'].startswith(k + '_'):
                f['action'] = True
                f['action_flag'] = True
                f['name'] = f['name'].split(k + '_', 1)[1]
                p += 1
    # if afp:
    #     fs = r['flags']
    #     l = []
    #     m = {}
    #     for a in afp:
    #         mod = a[1]
    #         action = a[2]
    #         main = a[3]
    #         f = fs[mod].pop(a[0] + m.get(mod, 0))
    #         m[mod] = m.setdefault(mod, 0) - 1
    #         l.append(f)
    #         if main:
    #             f['action'] = True
    #             f['default'] = 'ACTION'
    #         else:
    #             f['name'] = f['name'].split(action + '_', 1)[1]
    #     fs['actions'] = l
    return r


invisible_sep = u'\u2063'


def get_argv_val(k, ign_val=''):
    a = list(sys.argv)
    while a:
        v = a.pop(0)
        if v.startswith(k):
            if '=' in v:
                return v.split('=', 1)[1]
            else:
                if ign_val and a and a[0].startswith(ign_val):
                    return ''
                return a.pop(0) if a else ''


# coloring themeable,i.e. only using the base colors which themes modify:
ansi_col = lambda nr: '\033[%sm' % nr
col = lambda c: ansi_col(
    {
        'name': '1;33',
        'default': '0;31',
        'meaning': '0;38;5;245',
        'choices': '0;38',
        'details': '0;38;5;241',
        'short_name': '0;32',
        'action': '0;33',
        'default_action': '1;33',
    }.get(c)
)


def color_usage(*a, main_module, full=None, **kw):
    """
        colorizes flag keys and match from --help <match>
    """
    ret = []
    d = main_module.__doc__
    if d:
        ret.append('\n' + d.strip() + '\n\n')
    add = lambda s, end='\n': ret.append(s + end)
    # catch the original output:
    hof = get_argv_val('--help_output_fmt')
    hof = hof if hof else 'terminal'
    match = get_argv_val('--helpfull', ign_val='--help_output_fmt')
    match = [match] if match else []
    dmn = os.environ.get('doc_module_name')
    if dmn:
        match.append(dmn)

    # -h with an action flag? Then we detail its sub flags:
    afs = []
    for k in sys.argv:
        af = action_flags.get(k)
        if af:
            define_flags(af['flg_cls'], sub=af['key'], parent_autoshort=af['autoshort'])
            afs.append(af)

    so = sys.stdout
    s = io.StringIO()
    sys.stdout = s
    flags.FLAGS.write_help_in_xml_format(sys.stdout)
    sys.stdout = so
    u = s.getvalue()
    s.close()
    j = parse_xml_help(u, match=match, cli_actions=afs)
    all_flgs = j['flags']
    if hof != 'terminal':
        r = []
        [r.extend(v) for v in all_flgs.values()]
        add(tabulate(r, headers='keys', tablefmt=hof))
        return ret

    # only -h was given (artifically added match) but no real match - output as if
    # no match was given, the list is filtered at this point:
    if dmn and match == [dmn]:
        match = []
    match_hilite = ansi_col('1;32') + ' '.join(match) + ansi_col('0')
    j['match'] = '[matching %s]' % match_hilite if match else ''

    if match:
        n = 'All supported' if full else 'Main'
        add('%s command line flags %s:' % (n, j['match']))
    if hof != 'terminal':
        r = tabulate()

    def do(fn_mod):
        # module headline:
        flgs = all_flgs[fn_mod]
        if len(all_flgs) > 1:
            add(ansi_col('1;34') + fn_mod)
        w = term_widths(match)
        add(to_terminal(flgs, w, match=match))

    # main_help = all_flgs.pop(main_module.__name__, [])
    # show non main module flags on helpfull:
    n = main_module.__name__
    if full:
        [do(fn_mod) for fn_mod in sorted(all_flgs) if not fn_mod in (n, 'actions')]
    if n in all_flgs:
        do(n)
    if all_flgs.get('actions'):
        do('actions')

    # show main module's flags last - always:
    add('\033[0m')
    return ret


hlp_flags = {'-h': False, '--help': False, '-hf': True, '--helpfull': True}


def find_module(main):
    f, kw1 = func_from_partial(main)
    if f:
        return sys.modules.get(f.__module__)


def exit_at_help_flag(main, argv):
    """Called always by devapp.app at runtime start"""
    full, mod = None, None
    for a in argv:
        if a in hlp_flags:
            full = hlp_flags[a]
            break
    if full is None:
        return

    mod = find_module(main)
    if not mod:
        print('No module for function found: %s' % main)
        sys.exit(1)
    if a != '--helpfull':
        argv[argv.index(a)] = '--helpfull'

    h = color_usage(main_module=mod, full=full)
    print(''.join(h))
    sys.exit(0)



import traceback


def stripped_traceback(
    nostrip=0,
    formatted_exc=None,
    zope_match='Zope',
    zope_match2='zope',
    monkey_match='in monkey_exec',
    logging_exec_match='logging_exec',
):
    """ remove internals from a traceback.format_exc() structure
    Default is to remove Zope internals.
    """
    formatted_exc = formatted_exc or traceback.format_exc()
    tbr = formatted_exc.splitlines()
    if nostrip:
        return tbr
    tb = []
    i = -1
    while i < len(tbr) - 1:
        i += 1
        line = tbr[i]
        if line.endswith(monkey_match):
            i += 3
            continue

        if zope_match in line or zope_match2 in line or logging_exec_match in line:
            if tb[-1] != '    internal...':
                tb.append('    internal...')
            i += 1
            continue

        # good one append it:
        tb.append(line)
    # strip the v1 error handler itself from the traceback:
    if len(tb) > 3 and tb[-2].endswith('in err'):
        tbr = tb[:-3]
    else:
        tbr = tb
    return tbr



import time
from threading import current_thread

import structlog

from .common import transient_data_key


def positional_args_into_msg(logger, level, ev_dict):
    '.. but only if wanted...'
    pa = ev_dict.get('positional_args', None)
    if not pa:
        return ev_dict
    ev = ev_dict.get('event')
    if ev == '%s':
        ev_dict['event'] = ' '.join(str(s) for s in pa)
        del ev_dict['positional_args']
    else:
        try:
            if len(pa) == 1 and isinstance(pa[0], dict) and pa[0]:
                pa = pa[0]
            ev_dict['event'] = ev % pa
            del ev_dict['positional_args']
        except:
            pass
    return ev_dict


def censor_values(L, l, ev_dict, cfg):

    keys = cfg.get('keys')
    if not keys:
        return ev_dict

    for k in ev_dict:
        if k in keys:
            show_chars = cfg.get('hint_chars')
            v = cfg.get('repl_with', '*CENSORED*')
            if show_chars:
                v += ' hint: ' + str(unicode(ev_dict[k])[:show_chars]) + '...'
            ev_dict[k] = v
    return ev_dict


def selected_processors(L, l, ev_dict, cfg):
    """The processor function run as partial possibly a few times in the
    main proc chain. See config.py.
    """
    selector = cfg['selector']
    position = cfg['position']  # e.g. 'after_TimeStamper', just an index
    if selector == 'name':
        procs = L.ax_structlog_selected_processors.get(position)
        if procs:
            # maybe the proc wants to know its state.
            # we cannot change the sig, we have sl internal procs as well:
            ev_dict['_processor_state'] = cfg

            for p in procs:
                ev_dict = p(L, l, ev_dict)

            ev_dict.pop('_processor_state', 0)
    return ev_dict


def add_transient_data_holder(L, l, ev_dict):
    ev_dict[transient_data_key] = {}
    return ev_dict


def remove_transient_data(L, l, ev_dict):
    ev_dict.pop(transient_data_key, 0)
    return ev_dict


now = lambda: int(time.time() * 1000)
t0 = now()


def add_dt(_, __, ev):
    ev['timestamp'] = now() - t0
    return ev


def TimeStamper(**kw):
    timefmt = kw.get('fmt', 'ISO')
    if timefmt == 'dt':
        return add_dt
    utc = kw.get('utc', True)
    return structlog.processors.TimeStamper(timefmt, utc=utc)


def add_thread_name(L, l, ev_dict):
    t = current_thread()
    try:
        tn = int(t.getName().rsplit('-', 1)[1])
    except:
        if t.getName() == 'MainThread':
            tn = 0
        else:
            tn = t.ident  # int guaranteed
    tn = tn % 100000
    ev_dict['thread'] = tn
    # ev_dict['thread'] = 0
    return ev_dict



"""
Various Application Adapters
"""

import os

from theming.absl_color_help import setup_colorized_help


def adapt_gevent_pywsgi(server, logger):
    """
    Make gevent's pywsgi log to structlog.

    server = log_to_structlog(
         gevent.pywsgi.WSGIServer((host, port)), logger=structlog.get_logger('..')
    )
    """

    def log_4_gevent_pywsgi(self, logger=logger):
        """(Pdb) self
        <gevent.pywsgi.WSGIHandler object at 0x7f482697ba20>
        """
        length = self.response_length or '-'
        if isinstance(self.client_address, tuple):
            ip = {'ip': self.client_address[0], 'port': self.client_address[1]}
        else:
            ip = {'ip': self.client_address or '-'}
        if self.time_finish:
            ip['dt'] = int((self.time_finish - self.time_start) * 1000)
        logger.info(
            self.requestline or '',
            length=self.response_length or '-',
            status=(self._orig_status or self.status or '000').split()[0],
            **ip
        )

    server.handler_class.log_request = log_4_gevent_pywsgi
    return server


def absl_run(mod_name, main, write_pid=True):
    def do_write_pid():
        try:
            fn = FLAGS.pidfile
        except:
            # only when he gave the flag
            return
        os.makedirs(os.path.dirname(fn), exist_ok=True)
        with open(fn, 'w') as fd:
            fd.write(str(os.getpid()))

    # legacy call, use devapp runner

    setup_colorized_help(main)
    if write_pid:
        do_write_pid()
    app.run(main)


# .



"""
name="hurry.filesize",
version="0.9",
description="A simple Python library for human readable file
            sizes (or anything sized in bytes).",
classifiers=[
    "Programming Language :: Python",
    "Topic :: Software Development :: Libraries :: Python Modules",
    ],
keywords='file size bytes',
author='Martijn Faassen, Startifact',
author_email='faassen@startifact.com',
url='',
license='ZPL 2.1',


"""
traditional = [
    (1024 ** 5, 'P'),
    (1024 ** 4, 'T'),
    (1024 ** 3, 'G'),
    (1024 ** 2, 'M'),
    (1024 ** 1, 'K'),
    (1024 ** 0, 'B'),
]

alternative = [
    (1024 ** 5, ' PB'),
    (1024 ** 4, ' TB'),
    (1024 ** 3, ' GB'),
    (1024 ** 2, ' MB'),
    (1024 ** 1, ' KB'),
    (1024 ** 0, (' byte', ' bytes')),
]

verbose = [
    (1024 ** 5, (' petabyte', ' petabytes')),
    (1024 ** 4, (' terabyte', ' terabytes')),
    (1024 ** 3, (' gigabyte', ' gigabytes')),
    (1024 ** 2, (' megabyte', ' megabytes')),
    (1024 ** 1, (' kilobyte', ' kilobytes')),
    (1024 ** 0, (' byte', ' bytes')),
]

iec = [
    (1024 ** 5, 'Pi'),
    (1024 ** 4, 'Ti'),
    (1024 ** 3, 'Gi'),
    (1024 ** 2, 'Mi'),
    (1024 ** 1, 'Ki'),
    (1024 ** 0, ''),
]

si = [
    (1000 ** 5, 'P'),
    (1000 ** 4, 'T'),
    (1000 ** 3, 'G'),
    (1000 ** 2, 'M'),
    (1000 ** 1, 'K'),
    (1000 ** 0, 'B'),
]


def size(bytes, system=traditional):
    """Human-readable file size.

    Using the traditional system, where a factor of 1024 is used::
    
    >>> size(10)
    '10B'
    >>> size(100)
    '100B'
    >>> size(1000)
    '1000B'
    >>> size(2000)
    '1K'
    >>> size(10000)
    '9K'
    >>> size(20000)
    '19K'
    >>> size(100000)
    '97K'
    >>> size(200000)
    '195K'
    >>> size(1000000)
    '976K'
    >>> size(2000000)
    '1M'
    
    Using the SI system, with a factor 1000::

    >>> size(10, system=si)
    '10B'
    >>> size(100, system=si)
    '100B'
    >>> size(1000, system=si)
    '1K'
    >>> size(2000, system=si)
    '2K'
    >>> size(10000, system=si)
    '10K'
    >>> size(20000, system=si)
    '20K'
    >>> size(100000, system=si)
    '100K'
    >>> size(200000, system=si)
    '200K'
    >>> size(1000000, system=si)
    '1M'
    >>> size(2000000, system=si)
    '2M'
    
    """
    for factor, suffix in system:
        if bytes >= factor:
            break
    amount = int(bytes / factor)
    if isinstance(suffix, tuple):
        singular, multiple = suffix
        if amount == 1:
            suffix = singular
        else:
            suffix = multiple
    return str(amount) + suffix



"""
Tools which are importable before app runs
"""


import os
import platform
import sys
from getpass import getpass

from devapp.tools import write_config_cls_key

try:
    from devapp.app import app

    app.die
except:
    app = None

env = os.environ
ls = os.listdir


def j(*f):
    return os.path.join(*f)


def die(msg, **kw):
    app and app.die(msg, **kw)
    print(msg, str(kw))
    sys.exit(1)


def confirm(msg, default=True):
    y = 'Y' if default else 'y'
    n = 'n' if default else 'N'
    print('Confirm: %s [%s/%s] ' % (msg, y, n))
    try:
        v = input('> ')
    except KeyboardInterrupt as ex:
        print('Interrupted - bye.')
        sys.exit(1)
    except EOFError as ex:
        print('Interrupted - bye.')
        sys.exit(1)
    if not v:
        v = str(default)
    if v.lower() in ('y', 'yes', '1', 'true'):
        return True
    print('Unconfirmed')
    return False


def env_get_interactive_write(k, validate=None, example=None):
    v = env.get(k)
    if v is not None:
        return v
    msg = 'Required environ missing.'
    if example:
        msg += '\nExample value: %s\n' % example
    ask = getpass if 'assw' in k else input
    try:
        v = ask(msg + '\nEnter value for %s: ' % k)
    except KeyboardInterrupt as ex:
        v = 'q'
    except EOFError as ex:
        v = 'q'
    if v == 'q':
        print('Interrupted - bye.')
        sys.exit(1)
    if validate is not None:
        rv = validate(k, v)
        if rv == True:
            return v
        elif rv == 'write':
            write_config_cls_key(k, v)
            return v
        elif rv:
            print(rv)
        return env_get_interactive_write(k, validate, example)
    return v


def source2(fn=None, src=None, repl_ctx=None):
    m = {}
    if fn:
        if not os.path.exists(fn):
            raise Exception(fn, 'not found')
        with open(fn) as fd:
            s = fd.read()
    else:
        s = src
    if repl_ctx and '%(' in s:
        s = s % repl_ctx
    for l in s.splitlines():
        if l.strip() and '=' in l and not l.startswith('#'):
            k, v = l.split('=', 1)
            if v:
                if v[0] in ('"', "'"):
                    v = v[1:-1]
                m[k] = v
    return m


def get_arch():
    arch = platform.architecture()[0]
    if '64' in arch:
        arch = 'x86_64'
    elif '32' in arch:
        arch = 'x86_32'
    else:
        arch = 'unknown'
    return '%s-%s' % (platform.system(), arch)



from cgi import escape
from xml.dom import minidom


class BLESS_TERM:
    """Erradication the blessterm dependency which only did some ansi colors"""

    @staticmethod
    def color(color, s):
        return '\x1b[38;5;%sm%s\x1b(B\x1b[m' % (color, s)

    @staticmethod
    def nofmt(s):
        return s


#  term -> tag -> 8 means: format a tag to terminal color nummer 8.
# ax_xml __main__ prints colors all out:
TR069_STYLE = {
    'client': 'tr069_client',
    'markup': {
        'term': {
            'tag': 8,
            'text': 2,
            'attr': 0,
            'noxml': 12,
            'attr_val': 8,
            'tr_node': 8,
            'tr_leaf': 5,
            'tr_serial': 1,
        },
        'html': {
            'tag': 'rgb(142, 162, 162)',
            'text': '#00F',
            'attr': '#aaa',
            'text_bold': 1,
            'tr_leaf_bold': 1,
            'noxml': 'rgb(92, 124, 132)',
            'attr_val': '#aaa',
            'tr_node': 'rgb(104, 107, 200)',
            'tr_leaf': '#rgb(240, 232, 211)',
            'tr_serial': 'rgb(228, 50, 0)',
        },
    },
}


DEF_STYLES = {
    'client': '',
    'markup': {
        'term': {'noxml': 0, 'tag': 8, 'text': 2, 'attr': 0, 'noxml': 0, 'attr_val': 8},
        'html': {
            'indent': 2,
            'tag': '#8bd124',
            'text': '#000',
            'attr': '#555',
            'noxml': '#999',
            'attr_val': '#355',
        },
    },
}


def get_tag_val(tag, xml_str, match=None):
    """ quick access to tag values via string splitting
    FIXME: use regex...
    """

    def get_tag_content(stri):
        # string starts within a tag def, like ab='23' c='2'>foo</tag>
        # we return 'foo'
        return stri.split('</', 1)[0].split('>', 1)[1]

    if not match:
        return get_tag_content(xml_str.split('<%s' % tag, 1)[1])
    for tag_line in xml_str.split('<%s' % tag):
        tag_line = tag_line.split('</%s' % tag)[0]
        if match in tag_line:
            return get_tag_content(tag_line)


def pretty_print(xmlstr, style_overlay={}):
    """
    Called from outside, with a xml string given.

    Formats it to formats, like html or term.

    mode destines the formatting, we just search for
    control statements within it, like indent or compress
    """
    style = {}
    style.update(DEF_STYLES)
    style.update(style_overlay)
    # 1 or nothing:
    col = style.get('col')
    # 1 or nothing:
    do_indent = style.get('indent', 0)
    # semi or full or nothing:
    compress = style.get('compress')
    # html or term:
    format = style.get('format')
    if col:
        style['col_format'] = format

    try:
        el = minidom.parseString(xmlstr)
    except:
        # we should return sth - so...make it xml(text):
        xmlstr = str(xmlstr)
        if not col:
            return xmlstr
        res = col_format(xmlstr, '-', 'noxml', style)
        return escape_and_repl_html(res, style)

    if format == 'term':
        indent = int(do_indent) * ' '
        nl = '\n'
        if compress == 'full':
            nl = ''

    elif format == 'html':
        indent = int(do_indent) * '_AX_SP_'
        nl = '_AX_NL_'
        if compress == 'full':
            nl = ''
    else:
        return xmlstr

    if compress == 'semi':
        # need to pass this to the writexml
        # so that it makes a newline only at tag starts:
        style['semicompress'] = nl
        nl = ''

    myel = MyElement(el.firstChild, (col_format, style))
    # this is in minidom a function calling def writexml, which we overwrote below:
    res = myel.toprettyxml(indent=indent, newl=nl)

    # stupid html backreplacements:
    if format == 'html':
        res = escape_and_repl_html(res, style)
    return res


def escape_and_repl_html(res, style):
    # the stupid tag rewriting is to not have the writer replace it while doing toprettyxml :-/
    # replace is 2x faster then re.sub:
    res = escape(res)
    res = (
        res.replace('_AXO_', '<font color="')
        .replace('_AXC_', '">')
        .replace('_AXCC_', '</font>')
    )
    res = res.replace('_AX_SP_', '&nbsp;').replace('_AX_NL_', '<br>\n')
    res = res.replace('_AX_B_', '<b>').replace('_AX_BC_', '</b>')
    return res


def col_format(s, el, mode='tag', style={}):
    # either html or term:
    format = style.get('col_format', '')
    markup = style['markup'].get(format)
    if not markup:
        return s
    col = markup[mode]
    if mode == 'text' or mode == 'noxml':
        # keep newlines:
        if format == 'html':
            s = s.replace('\n', '_AX_NL_')
    try:
        res = check_special_clients(s, el, mode, style)
        if res:
            return res
    except:
        pass

    if format == 'html':
        if markup.get(mode + '_bold'):
            s = '%s%s%s' % ('_AX_B_', s, '_AX_BC_')
        return '_AXO_%s_AXC_%s_AXCC_' % (col, s)

    elif format == 'term':
        return BLESS_TERM.color(col, s)


def check_special_clients(s, el, mode, style):
    """ that's the main reason why we do this mess instead of
    usual xml formatting via xsl bla: """
    # special handling for some prominent clients:
    # those things are the reason y i don't use xslt:
    # the tr-069 command line client:
    if 'cwmp:ID' in str(el):
        pass
    if style.get('client', '') == 'tr069_client':
        if mode == 'tag' and s.startswith('<cwmp:'):
            begin, end = s.split(':', 1)
            if not end:
                return
            return '%s%s' % (
                col_format(begin + ':', el, mode, style),
                col_format(end, el, 'tr_leaf', style),
            )
        if not mode == 'text':
            return

        if el.tagName == 'Name':
            # coloring just the leafs of a tr-069 name:
            begin, end = s.rsplit('.', 1)
            return '%s.%s' % (
                col_format(begin, el, 'tr_node', style),
                col_format(end, el, 'tr_leaf', style),
            )
        elif el.tagName == 'SerialNumber':
            return col_format(s, el, 'tr_serial', style)


class MyElement(minidom.Element):
    """ pretty much the original, except the format calls """

    def __init__(self, el, fmt):
        self.tagName = getattr(el, 'tagName', None)
        self.nodeName = getattr(el, 'nodeName', None)
        self.prefix = el.prefix
        self.namespaceURI = el.namespaceURI
        self.childNodes = el.childNodes
        self._attrs = getattr(el, '_attrs', {})
        self._attrsNS = getattr(el, '_attrsNS', {})
        self.fmt, self.style = fmt

    def writexml(self, writer, indent='', addindent='', newl=''):
        """ this goes plain stupid through an xml tag with its attrs and text"""
        style = self.style
        close_tag = self.fmt('>', self, 'tag', style)
        # indent = current indentation
        # addindent = indentation to add to higher levels
        # newl = newline string
        writer.write(
            style.get('semicompress', '')
            + indent
            + self.fmt('<' + self.tagName, self, 'tag', style)
        )

        attrs = self._get_attributes()
        a_names = list(attrs.keys())
        a_names.sort()

        for a_name in a_names:
            # attribute names:
            writer.write(' %s' % self.fmt(' %s="' % a_name, self, 'attr', style))
            minidom._write_data(
                writer, self.fmt(attrs[a_name].value, self, 'attr_val', style)
            )
            writer.write(self.fmt('"', self, 'attr', style))

        if self.childNodes:
            if (
                not newl
                and len(self.childNodes) == 1
                and self.childNodes[0].nodeType == minidom.Node.TEXT_NODE
            ):
                writer.write(
                    '%s%s%s%s'
                    % (
                        close_tag,
                        self.fmt(self.childNodes[0].data, self, 'text', style),
                        self.fmt('</%s>' % self.tagName, self, 'tag', style),
                        newl,
                    )
                )
                return

            ind = ''
            if newl:
                ind = indent + addindent
            writer.write('%s%s' % (close_tag, newl))
            for node in self.childNodes:
                if node.nodeName == '#text':
                    if node.data.strip():
                        minidom._write_data(
                            writer,
                            '%s%s%s'
                            % (ind, self.fmt(node.data, self, 'text', style), newl),
                        )
                else:
                    MyElement(node, (self.fmt, self.style)).writexml(
                        writer, indent + addindent, addindent, newl
                    )
            writer.write(
                '%s%s%s%s'
                % (
                    style.get('semicompress', ''),
                    indent,
                    self.fmt('</%s>' % self.tagName, self, 'tag', style),
                    newl,
                )
            )
        else:
            writer.write('%s%s' % (self.fmt('/>', self, 'tag', style), newl))


if __name__ == '__main__':
    print('term colors')
    if BLESS_TERM:
        for i in range(0, 255):
            print(BLESS_TERM.color(i, str(i)))

    xml_str = "<doc><foo bar='1'>tag val</foo><foo class='err'><innerfoo>tag err</innerfoo></foo></doc>"

    print((get_tag_val('foo', xml_str)))
    print((get_tag_val('foo', xml_str, 'err')))
    print()
    print(('Pretty printing this: %s' % xml_str))
    print(('-' * 100))
    for style in [
        {'format': 'html', 'col': 1},
        {'format': 'html', 'compress': 'full', 'col': 1},
        {'format': 'term', 'col': 1, 'compress': 'full', 'indent': 1},
        {'format': 'term', 'col': 1, 'indent': 1, 'compress': 'semi'},
    ]:
        print(('.' * 100))
        print(('Style %s:' % style))
        print(('.' * 100))
        print((pretty_print(xml_str, style_overlay=style)))
        print((pretty_print('noxml', style_overlay=style)))



def to_md_list(shorts, keys, funcs, details):
    """helper for max explain and absl flags parser"""
    # if 'temporary' in str(funcs):
    #    breakpoint()  # FIXME BREAKPOINT
    if not shorts:
        return ''
    sep = ' `|` ' if any([s for s in shorts if s]) else ' '
    w = max([len(k) for k in keys])
    ws = max([len(k) for k in shorts])
    out = []
    for s, k, f, det in zip(shorts, keys, funcs, details):
        if f.startswith('[') and ']' in f:
            d, f = f.split(']', 1)
        f = d + ']`%s`' % f  # low color
        out.append('* %s%s%s`:` %s' % (s.ljust(ws), sep, k.ljust(w), f))
        if det:
            out[-1] += ' *Details: %s*' % det

    out.append('')
    return '\n'.join(out)


import os

import yaml


def add_metas(md_or_file, metas):
    if isinstance(metas, dict):
        metas = [(k, v) for k, v in metas.items()]

    if os.path.exists(md_or_file):
        with open(md_or_file) as fd:
            s = fd.read()
    else:
        s = md_or_file

    sep, m, body = get_meta(s)
    if not body:
        s = '\n'.join(('---', '', '---', '')) + s
    sep, m, body = get_meta(s)

    for k, v in metas:
        m[k] = v
    ls = '\n\n'
    return ''.join(
        (
            sep,
            ls,
            yaml.safe_dump(m, default_flow_style=False).strip(),
            ls,
            sep,
            ls,
            body.strip(),
        )
    )


def get_meta(s):
    try:
        sep = s.strip().split('\n')[0]
        s = '\n' + s
        l = s.split('\n%s\n' % sep, 3)
        assert not l[0].strip()
        if not l[1].strip():
            m = {}
        else:
            m = yaml.safe_load(l[1])
        return sep, m, l[2]
    except:
        return None, None, None


def struct_code_block(d, title=None):
    r = '\n```yaml\n%s\n```\n\n' % yaml.dump(d, default_flow_style=False)
    if not title:
        return r
    s = '''
<details>
 <summary>%s</summary>
%s
</details>
    '''
    return s % (title, r)


if __name__ == '__main__':
    for md in (
        '#foo\nbody',
        '''
---
foo: bar
bar: baz
---

body
    ''',
        'foo',
    ):
        m = {'dict': {'a': 'b'}}
        md = add_metas(md, m)
        import os

        m['Env'] = dict(os.environ)
        md = add_metas(md, m)
        print(add_metas(md, (('a', 'b'),)))


# .



#!/usr/bin/env python
# coding: utf-8

import os
import sys

from tree_builder import *

gitl_url = 'https://gitlab-runner-downloads.s3.amazonaws.com/latest/binaries/gitlab-runner-linux-amd64'
hugo_url = 'https://github.com/gohugoio/hugo/releases/download/v0.59.0/hugo_0.59.0_Linux-64bit.tar.gz'
plnt_url = 'https://netcologne.dl.sourceforge.net/project/plantuml/plantuml.jar'


class Ports:
    gitlab = Env('DA_PORT_GITLAB', 5378)


class Project(T):
    'Gitlab'


class Service(T):
    runtime = 'systemd'
    service_restart_sec = 2


build_type_hirarchy(root=Project)


class GitlabRunner(Service):
    hugo_dir = Env('DA_DIR') + '/hugo'
    themes_dir = hugo_dir + '/themes'
    filesystem = [
        {
            'name': 'gitlab',
            'type': 'conda_env',
            'packages': ['python', 'java-1.7.0-openjdk-devel-cos6-x86_64'],
        },
        {'name': 'gitlab_runner', 'type': 'exe', 'url': gitl_url, 'fs_method': 'cp',},
        {
            'name': 'https://github.com/matcornic/hugo-theme-learn.git',
            'checkout_dir': themes_dir + '/hugo-theme-learn',
        },
        {
            'name': 'hugo',
            'type': 'exe',
            'url': hugo_url,
            'fs_method': 'cp',
            'checkout_dir': hugo_dir,
            'checkout_have_check_dir': hugo_dir + '/bin',
        },
        {
            'name': 'plantuml.jar',
            'type': 'lib',
            'url': plnt_url,
            'fs_method': 'cp',
            'checkout_dir': hugo_dir + '/lib',
        },
    ]
    CONFIG_FILE = '%(etc_dir)s/runner.toml'
    descr = 'devapps'
    tags = ['devapps', 'production']
    app = 'gitlab_runner'
    app_args = ['run']
    # shortcut for:
    # service_exec_start = [app, 'run'] + _cfg
    service_kill_mode = 'process'
    url = 'https://gitlab.my_company.com'
    java = '%(DA_DIR)s/envs/gitlab/x86_64-conda_cos6-linux-gnu/sysroot/usr/lib/jvm/java-1.7.0-openjdk-1.7.0.131.x86_64/bin/java'

    # token = 'mFsHPcAsRKhpsnstXMVV'

    functions = {
        'java': [java],
        'plantuml': [java, '-jar', hugo_dir + '/lib/plantuml.jar', '-tsvg'],
        'plantrender': ['%(DA_DIR)s/conf/SPEC.py', 'render_plantuml'],
        'register': [
            app,
            'register',
            '--non-interactive',
            '--url',
            url,
            '--registration-token',
            Env(
                'registration_token',
                suggest=[
                    Env('HOME') + '/.config/devapps/creds',
                    'gitlab_runner_reg_token',
                ],
            ),
            '--executor',
            'shell',
            '--description',
            descr,
            '--tag-list',
            ','.join(tags),
            '--locked',
            'false',
        ],
        'unregister': [app, 'unregister', '--all-runners'],
    }
    service_exec_start_pre = 'register'
    # service_exec_stop_post_1 = 'unregister'


class Gitlab(Project, GitlabRunner):
    ''


root = create_tree(Gitlab)


args = sys.argv


def render_plantuml(dir_):
    """
    Replacing `` `plantuml fenced blocks with rendered svgs, inline

    - Intended to be run before hugo generating gitlab pages (i.e. within .gitlab-ci.yml)
    - Requires plantuml
    - Run via e.g. $app_run_script_path plantrender /home/gk/repos/axwifi-server/docs/content/
    - To be replaced with a rmarkdown function, once we have R-Merkdown
    """

    def repl(fn):
        with open(fn) as fd:
            s = fd.read()
        arsp = os.environ['app_run_script_path']
        parts = s.split('\n```plantuml\n')
        print('Replacing %s plantumls within %s' % (len(parts) - 1, fn))
        r = parts[0]
        for part in parts[1:]:
            uml, rest = part.split('\n```\n', 1)
            cmd = "echo '%s' | '%s' plantuml -tsvg -p " % (uml, arsp)
            svg = os.popen(cmd).read().split('<svg', 1)[1]
            svg, src = svg.split('<!--', 1)
            src = src.split('@startuml', 1)[1].split('@enduml', 1)[0]
            uml = '<svg%s</g></svg>' % svg
            r += '<div style="background: #eeeeFF">%s</div>' % uml
            r += (
                '<details><summary>Source</summary>\n\n```uml\n@startuml\n%s\n@enduml\n```\n</details>\n%s'
                % (src, rest)
            )
        with open(fn, 'w') as fd:
            fd.write(r)

    os.chdir(dir_)
    mds = set(
        [f.split(':', 1)[0] for f in os.popen('grep -ir plantuml').read().splitlines()]
    )
    [repl(f) for f in mds]


if __name__ == '__main__':
    if args[1] == 'render_plantuml':
        sys.exit(render_plantuml(args[2]))

    from tree_builder.links import dump_all

    print(dump_all(root, fmt='json'))



#!/usr/bin/env python
"""
This is a working literal translation of the jq based moby-download frozen
image tool.


Could be done far smaller.

"""

import os
import signal
import sys
import time

from devapp.app import flag, run_app

flag.string('dir', './images', 'Exisiting target dir', short_name='d')
flag.string('repo', 'busybox:latest', 'repo')


def cleanup(*args):
    print('Exiting')
    os.system('touch /root/foooo')
    sys.exit(0)


signal.signal(signal.SIGINT, cleanup)
signal.signal(signal.SIGTERM, cleanup)


def main():
    for i in range(1, 100):
        print('stasring')
    while True:
        time.sleep(60)


run = lambda: run_app(main)

if __name__ == '__main__':
    run()



try:
    __import__('pkg_resources').declare_namespace(__name__)

except ImportError:

    from pkgutil import extend_path

    __path__ = extend_path(__path__, __name__)


"""
This is ususally imported only when logging is configured.
Our features depend on structlog which is not a dependency.
"""
try:
    pass
except ImportError as ex:
    ex.args += (
        'Hint: You have to add structlog to your packages.' '=> pip install structlog',
    )
    raise


# -- Setup for a stdlib logging free, getattr free use:
